/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/

#ifndef CONSTS_H
#define CONSTS_H

#define EXIT_RETURN_SUCCESS 0
#define EXIT_RETURN_FAILURE 1

#ifndef YES
#define YES 1
#define NO  0
#endif

#define INITIAL_MESH_QUALITY_SIGMA 6.0        /* */
#define INITIAL_MESH_QUALITY_OMEGA 1.5        /*   mesh-quality params */
#define INITIAL_MESH_QUALITY_TAU   3.0        /* */ 
#define INITIAL_MERGE_TOLERANCE    1.e-5      /* merge tolerance */
#define INITIAL_CELL_SIZE_TO_CDIM  6.0        /* ratio of cell size to cdim */
                                              /* ratio of remap trigger */
#define INITIAL_CONSTMCDIM         0.0        /* initial constant mesh-size */
#define INITIAL_EXPONENT           2.0        /* initial mesh-control */
					      /* function exponent */

#ifndef UNIX
#define LOG_BETA (0.064850972) /* LOG( BETA ); 
      BETA = 1.2  --> LN(1.2)  = 0.1823215932 
      BETA = 1.1  --> LN(1.2)  = 0.9531020373E-01
      BETA = 1.055--> LN(1.055)= 0.053540767
      BETA = 1.067--> LN(1.067)= 0.064532271
      BETA = 1.05 --> LN(1.05) = 0.4879011959E-01     
      */
#define ONE_OVER_LOG_BETA 15.419969 /* 1. / LOG_BETA  */
#else
#   define INITIAL_BETA 1.05
#endif

#define INITIAL_MAGFACTOR 1.1

#define SAFETY_MARGIN 10.0  /* to set up the array into which
                              the hash function shoots */


#define NO_CDIM_AVAILABLE 0.0  /* value used as argument to ADDVERTEX */

#define MINIMAL_SIDE_LENGTH 0.60   /* value used as lower bound on the
                                      stretch value in computation of ideal
				      radius and ideal point */
#define MINIMAL_HEIGHT      0.3316 /* sqrt(MINIMAL_SIDE_LENGTH**2.-0.25) */

                              /* MINIMAL_HEIGHT/(sqrt(3.)*2.)  */
#define MINETA_FOR_EQUILATERAL (MINIMAL_HEIGHT/0.866025) 
       
#define ETASTEP_FOR_EQUILATERAL ((-(1.0 - MINETA_FOR_EQUILATERAL))*0.249)

#define ALMOST_ZERO  0.000000001  /* value used in EDGESINTERSECT */
#define ALMOST_ONE   0.999999999  /* value used in EDGESINTERSECT */

#define ALMOST_PARALELL(d) (fabs(d) < 1.e-30) /* value used in EDGESINTERSECT
					       to denote value of delta for
					       almost parallel lines */

#define LEFT_SHIFT  0.95 /* shifts used in IDEALPOINT to move the ideal point
			   in the direction of the pulled-edge */
#define RIGHT_SHIFT 0.05 /* i.e. (1.-LEFT_SHIFT) */   

#define DEFAULT_EXPONENT_OF_MCTL_FUNC  2

#define SHAPE_FUNC_VAL_IN_INFL_RADIUS 0.90 /* influence radius is defined */
					   /* as a radius within which the */
					   /* shape function takes values */
					   /* >= prescribed value. The */
					   /* prescribed value is just */
					   /* SHAPE_FUNC_VAL_IN_INFL_RADIUS */

#define SHAPE_FUNC_B   /* select type of shape function used in mesh-control */
                       /* another choice is SHAPE_FUNC_A */

#define LEAF_ARRAY_TOTAL 100. /* Problem dependent constant giving approximate
                                 number of leafarray matrices at the start of 
                                 meshing */

#define MIN_LEAF_DIM  (long)2  /* minimal dimension of the leaf arrays (>1) */

#ifdef UNIX
#define MAX_FRONT_LISTS  2000  /* dimension of the hash table */
#else
#define MAX_FRONT_LISTS  256
#endif

#define VERY_SMALL_NMBR 1.e-30
#define PI              3.1415927
#define PI_TIMES_2      6.2831853

#define RELAX_CONST 0.95
#define MAX_RELAX_PASSES 2

#define MAX_SMOOTH_PASSES 4

#define COS30DEG 0.866 
#define SMALLANGCOS COS30DEG 

#define  UPPER_LIMIT_ON_IDEAL_RADIUS 1.7
#define  LOWER_LIMIT_ON_IDEAL_RADIUS MINIMAL_SIDE_LENGTH 
#define  IDEAL_RADIUS_REDUCTION 1.0
#define  SEARCH_RADIUS_REDUCTION 1.0

#define SIZE_FRACT(d) 0.25*d /* ratio of size of mctl triangle to cdim  */

#ifndef FLT_MAX
#define FLT_MAX 3.40282347e+38F
#endif
#ifndef FLT_MIN
#define FLT_MIN 1.17549435e-38F
#endif

#ifdef ULONG_MAX
#   define MAX_CELLS 	ULONG_MAX /* 2^32 -- a limit to avoid overflow */
#else
#   define MAX_CELLS    4294967295 
#endif

#define NO_SPECIAL_VERTEX_ID 0 

#define SQRT_2 1.4142136

#define ONE_THIRD 0.333333333333333333333333333333

#undef BY_GRADIENT  /* BY_GRADIENT gives by about 50% less quadrants
			than the gradient-insensitive method (for
			BY_GRADIENT_SIZE_FRACTION = 1). On the other hand
			the generated meshes are slightly worse looking. */

#define BY_GRADIENT_SIZE_FRACTION 0.5

#define Z_SHOW_SCALE 0.10

#define ANISOTROPIC_SMOOTHING 1
#define LAPLACIAN_SMOOTHING   2


#endif
