/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/
 
#ifndef CURVE_H
#define CURVE_H

#include "ttypes.h"
  
void make_curve_list(void);
CURVE_REC *add_curve_line(int id, float fx, float fy, float lx, float ly);
CURVE_REC *add_curve_discretized(int id, float fx, float fy,
		      float lx, float ly);
CURVE_REC *add_curve_circle(int id, float fx, float fy, float cx, float cy);
CURVE_REC *add_curve_circ_arc(int id, 
	       float fx, float fy,
	       float lx, float ly,
	       BOOLEAN intermed_p_is_valid,
	       float ix, float iy, 
	       float cx, float cy, BOOLEAN opposite_dir);

void interpolate_line(float eta, CURVE_REC *, float *, float *);
void interpolate_discretized(float eta, CURVE_REC *, float *, float *);
void interpolate_circle(float eta, CURVE_REC *, float *, float *);
void interpolate_circ_arc(float eta, CURVE_REC *, float *, float *);

void print_line(CURVE_REC *);
void print_discretized(CURVE_REC *);
void print_circle(CURVE_REC *);
void print_circ_arc(CURVE_REC *);

double length_line(CURVE_REC *);
double length_discretized(CURVE_REC *);
double length_circle(CURVE_REC *);
double length_circ_arc(CURVE_REC *);

void adj2grad_line(sense, sense, CURVE_REC *);
void adj2grad_discretized(sense, sense, CURVE_REC *);
void adj2grad_circle(sense, sense, CURVE_REC *);
void adj2grad_circ_arc(sense, sense, CURVE_REC *);

void extent_line(float *xmin, float *ymin, float *xmax, float *ymax,
		 CURVE_REC *c_p);
void extent_discretized(float *xmin, float *ymin, float *xmax, float *ymax,
			CURVE_REC *c_p);
void extent_circle(float *xmin, float *ymin, float *xmax, float *ymax,
		   CURVE_REC *c_p);
void extent_circ_arc(float *xmin, float *ymin, float *xmax, float *ymax,
		     CURVE_REC *c_p);

CURVE_REC  *find_curve_by_id(int curve_id);
int get_curve_id(CURVE_REC *c);
int mesh_curve(void *, void *);
void gen_edges_on_curve(CURVE_REC *c_p);
void vertex_spacing(CURVE_REC *c_p, double length, int *nvert);
float predict_cdim(float d1, float eta, CURVE_REC *c_p, double len1);
void update_by_curve_extent(CURVE_REC *c_p,
		      float *xmin, float *ymin, float *xmax, float *ymax);
void generate_edges_of_discretized_curve(CURVE_REC *c);

#endif
