/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/

#ifndef INPUT_H
#define INPUT_H

#include "ttypes.h" /* here are the typedefs used below */


typedef enum { NO_TOKEN, STRING, NUMBER } token_type;

typedef struct {
  int    token_type;
  char  *val;
} token_rec;

#define NO_TOKENS -1

typedef struct  {
  int             tokens_total; /* >= 1         for non-empty command line */
				/*  = NO_TOKENS for empty line */
  int              line_number;
  token_rec tokens[MAX_TOKENS]; /* values of the tokens */
} line_contents;


void        input_data(void);
int         read_next_line(void);
int         parse_line(char *);
void        init_command_line(void);
int         get_tokens_total(void);
token_type  get_token_type(int token_number);
char        *get_token_as_string(int token_number);
double      get_token_as_double(int token_number);
int         bad_token(int ith, char *should_be_type);
int         current_command(void);
int         is_keyword_in_line(int start_at, char *keyword);
BOOLEAN     compare_token_with_keyword(char *tok, char *keyword);

/* These macros should be used to access contents of the */
/* command line: they do error checking on token type. */
#define token_double_val(ith) (get_token_type(ith)==NUMBER)?\
                                     (double)get_token_as_double(ith):\
                                     (double)bad_token(ith,"NUMBER")

#define token_float_val(ith) (get_token_type(ith)==NUMBER)?\
                                     (float)get_token_as_double(ith):\
                                     (float)bad_token(ith,"NUMBER")

#define token_int_val(ith)    (get_token_type(ith)==NUMBER)?\
                                     (int)get_token_as_double(ith):\
                                     (int)bad_token(ith,"NUMBER")

#define token_string_val(ith) (get_token_type(ith)==STRING)?\
                                     (char *)get_token_as_string(ith):\
                                     (char *)bad_token(ith,"STRING")

void        print_line_contents(void);
void        bad_syntax(char *);

#endif 
