/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/

#ifdef Q_TREE

#ifndef QTREE_H
#define QTREE_H

#include "ttypes.h"
#include "mctlp.h"

#ifdef X_XFACE

#include "xiface.h"

#endif

struct _QTREE_REC;

typedef struct _QTREE_LEAF_REC {
  BOOLEAN    have_ctl_p;
  float       ctlp_at_x;
  float       ctlp_at_y;
} QTREE_LEAF_REC;

typedef struct _QTREE_BRANCH_REC {
  struct _QTREE_REC        *ll_leaf;
  struct _QTREE_REC        *lr_leaf;
  struct _QTREE_REC        *ur_leaf;
  struct _QTREE_REC        *ul_leaf;
} QTREE_BRANCH_REC;

typedef struct _QTREE_REC {
  union {
    QTREE_LEAF_REC            leaf;
    QTREE_BRANCH_REC        branch;
  } spec;
  float                   center_x;
  float                   center_y;
  float                  dimension;
  float                      c_dim; /* c_dim == 0 signals branch */
#ifdef ANISOTROPIC_MESH
  float                         mf;
  float                        ang;
#endif
#ifdef NEED_PARENT
  struct _QTREE_REC      *parent_q;
#endif
#ifdef X_XFACE
#ifdef NEED_QUAD_GREP
  GraphicObj               *g_rep;
#endif
#endif  
} QTREE_REC;

void 
qt_prepare_for_use(void);
float 
qt_leaf_cdim_at(float x, float y);
float 
qt_2nd_dir_magn_factor_at(float x, float y);
float 
qt_1st_dir_angle_at(float x, float y);
void 
qt_make_root(float minx, float miny, float maxx, float maxy);
void 
qt_extend_qtree(float x, float y, float dim, float mf, float ang);
void 
qt_build(void);
BOOLEAN 
qt_get_mctlp(float *x, float *y, float *d);
void 
qt_add_mctlp(float x, float y, float d);

#endif /* ifdef QTREE_H */

#endif /* ifdef Q_TREE */
