/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/

#ifndef TARGE2CFG_H
#define TARGE2CFG_H

/* The configure-generated file. */
#include <Tcfg.h>

#ifdef MSDOS

#      define SEP           "\\"
#      define IN_SUFFIX     ".in"
#      define OUT_SUFFIX    ".out"
#      define MOS_SUFFIX    ".mos"
#      define WASP_SUFFIX   ".wsp"
#      define EST_SUFFIX    ".est"

#      define STANDARD_MALLOC		/* Must be used under MSDOS */
#      define INPUT_COMPATIBILITY 1

#      undef MCTL_BY_FUNC 
#      undef X_XFACE
#      define M_PI 3.14159265358979
#      define M_PI_2 1.57079632679490
#      define STRICT_MESH_SIZE_CONTROL 1
#      define ANISOTROPIC_MESH 1

#      ifdef BUNDLE_OPTS

#         define FAST_ACCESS
#         define DOS_GXFACE
#         define INPUT_COMPATIBILITY

#      endif

#else  /* !MSDOS */

#   ifdef UNIX

#      define SEP           "/"
#      define IN_SUFFIX     ".in"
#      define OUT_SUFFIX    ".out"
#      define MOS_SUFFIX    ".mos"
#      define WASP_SUFFIX   ".wsp"
#      define EST_SUFFIX    ".est"

#      include <ckitcfg.h> /* To be consistent with the Ckit configuration. */
                           /* See the file and the README of Ckit. */
#      ifdef BUNDLE_OPTS

/* These options are bundled.  You can define them in the Imakefile, so */
/* that you can enable or disable them, overriding thus the settings    */
/* below.  To do this set                                               */
/*                  -DOPTION_YOU_WANT=1       and                       */
/*                  -DOPTION_YOU_DONT_WANT=0                            */

/* Effect:  Some function calls are replaced by macros.  Speeds up */
/*          the mesh generation quite a bit. */

#         ifdef FAST_ACCESS
#            if !(FAST_ACCESS)  /* not desired */
#               undef FAST_ACCESS
#            endif
#         else
#            define FAST_ACCESS /* bundled */
#         endif

/* Effect:  Enables use of the quadtree-based mesh control function. */

#         ifdef Q_TREE
#            if !(Q_TREE)  /* not desired */
#               undef Q_TREE
#            endif
#         else
#            define Q_TREE /* bundled */
#         endif

/* Effect:  Enables generation of anisotropic meshes (i.e. oriented, */
/*          stretched in presribed directions. */

#         ifdef ANISOTROPIC_MESH 
#            if !(ANISOTROPIC_MESH)  /* not desired */
#               undef ANISOTROPIC_MESH
#            endif
#         else
#            define ANISOTROPIC_MESH /* bundled */
#         endif

/* Effect:  Strict mesh size enforcement control. */

#         ifdef STRICT_MESH_SIZE_CONTROL
#            if !(STRICT_MESH_SIZE_CONTROL)  /* not desired */
#               undef STRICT_MESH_SIZE_CONTROL
#            endif
#         else
#            define STRICT_MESH_SIZE_CONTROL /* bundled */
#         endif

/* Effect:  Enables on-line display of the generation progress using */
/*          the X Window System and the ELIXIR library. */

#         ifdef X_XFACE 
#            if !(X_XFACE)  /* not desired */
#               undef X_XFACE
#            endif
#         else
#            define X_XFACE /* bundled */
#         endif

/* Effect:  Makes the input subsystem compatible with older versions */
/*          of the input specification. */

#         ifdef INPUT_COMPATIBILITY
#            if !(INPUT_COMPATIBILITY)  /* not desired */
#               undef INPUT_COMPATIBILITY
#            endif
#         else
#            define INPUT_COMPATIBILITY /* bundled */
#         endif

/* Effect:  Enables the use of quasi-estimator of the mesh size. */
/*          This is generally of use only in special cases (used */
/*          by the author to verify the generation procedure). */

#         ifdef QUASI_ESTIMATOR
#            if !(QUASI_ESTIMATOR)  /* not desired */
#               undef QUASI_ESTIMATOR
#            endif
#         else
#            undef QUASI_ESTIMATOR /* not desired */
#         endif

/* Effect:  Enables the use of mesh control by analytically defined */
/*          mesh size functions. */
/*          This is generally of use only in special cases (used */
/*          by the author to verify the generation procedure). */

#         ifdef MCTL_BY_FUNC
#            if !(MCTL_BY_FUNC)  /* not desired */
#               undef MCTL_BY_FUNC
#            endif
#         else
#            define MCTL_BY_FUNC /* bundled */
#         endif

/* Effect:  Disables drawing of generated triangles (only edges */
/*          are drawn in that case).  Speeds up interaction during */
/*          post-generation inspection (especially for large problems */
/*          of the order of 100000 and more triangles). */

#         ifdef DONT_DRAW_TRIANGLES 
#            if !(DONT_DRAW_TRIANGLES)  /* not desired */
#               undef DONT_DRAW_TRIANGLES
#            endif
#         else
#            define DONT_DRAW_TRIANGLES /* bundled */
#         endif


/* Effect:  Enables drawing of generated triangles as filled entities. */
/*          Degrades interactivity during */
/*          post-generation inspection (especially for large problems */
/*          of the order of 100000 and more triangles). */

#         ifdef FILL_TRIANGLES
#            if !(FILL_TRIANGLES)  /* not desired */
#               undef FILL_TRIANGLES
#            endif
#         else
#            undef FILL_TRIANGLES /* not desired */
#         endif

#      endif

/* The author is using error handlers during debugging as the default */
/* is on his machine to ignore silently some errors (e.g. division by zero). */
#      ifdef AIXV3
#         define CUSTOM_ERROR_HANDLER 
#         define IBM_ERROR_HANDLING
#      endif

#   else /* !UNIX */

You should have defined either MSDOS or UNIX. There is no other choice.

#   endif

#endif

#define NDEBUG 

#endif
