/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/

#ifndef VERTEX_H
#define VERTEX_H

#include "ttypes.h"

void       
make_vertex_list(void);
long int   
total_of_vertices(void);
VERTEX_REC *
vertex_near(float fx, float fy);
VERTEX_REC *
add_vertex(int id, float fx, float fy,
           float cdim, float mf, float ang);
void
destroy_vertex(VERTEX_REC *v_p);
float      
get_stored_cdim(VERTEX_REC *v1_p);
float      
get_stored_mf(VERTEX_REC *v1_p);
float      
get_stored_angle(VERTEX_REC *v1_p);
void       
set_vertex_curve_p(VERTEX_REC *v_p, CURVE_REC *c_p);
void       
get_vertex_coords(VERTEX_REC *v_p, float *x, float *y);
int        
add_vert_reference(VERTEX_REC *v_p, EDGE_REC *e_p);
int 
remove_vert_reference(VERTEX_REC *v_p, EDGE_REC *e_p);
void       
set_vert_curve_p(VERTEX_REC *v_p, CURVE_REC *c_p);
CURVE_REC  *
get_vert_curve_p(VERTEX_REC *v_p);
void 
select_fixed_point(float ipx, float ipy,
                   float fx, float fy, float lx, float ly,
                   float ecx, float ecy, float rad,
                   VERTEX_REC **selvertexp,
                   float *selvx, float *selvy);
void       adj_fixed_points(VERTEX_REC *selvertexp);
int        
get_vert_references(VERTEX_REC *v_p, int *n, EDGE_REC **ea_p);
void       
smooth_pass(int smoother_type);
void       
add_fixed_point(VERTEX_REC *p, int curve_id);
void       
dump_nodes_in_afro_style(FILE *fp);
void       
dump_nodes_in_mos_style(FILE *fp);
void       
vertex_data(VERTEX_REC *v, int *id, float *x, float *y);

int        
fixed_vertices_on_curve(CURVE_REC *c_p);
VERTEX_REC *sel_fixed_on_curve(float prevx, float prevy,
			       float *currx, float *curry);
void 
produce_mctlps(void);



/* MAKE ACCESSING OF VERTEX DATA FAST BY DEFINING MACROS THAT ARE USED */
/* INSTEAD OF FUNCTIONS */

#ifdef FAST_ACCESS

#ifndef VERTEX_PRIVATE_HDR
extern VERTEX_LIST vertex_list;
#endif

#define set_vertex_curve_p(aVP, c_p) \
              ((VERTEX_REC *)(aVP))->parent_curve = (CURVE_REC *)(c_p)
#define get_stored_cdim(aVP) (((VERTEX_REC *)(aVP))->stored_cdim)
#ifdef ANISOTROPIC_MESH
#   define get_stored_mf(aVP) (((VERTEX_REC *)(aVP))->stored_mf)
#   define get_stored_angle(aVP) (((VERTEX_REC *)(aVP))->stored_ang)
#else
#   define get_stored_mf(aVP)    (1)
#   define get_stored_angle(aVP) (0)
#endif
#define get_vertex_connectivity(aVP) ((aVP)->connectivity)
#define get_vertex_coords(aVP, ax, ay) \
              *((float *)(ax)) = ((VERTEX_REC *)(aVP))->coords.x; \
              *((float *)(ay)) = ((VERTEX_REC *)(aVP))->coords.y
#define set_vertex_coords(aVP, ax, ay) \
              ((VERTEX_REC *)(aVP))->coords.x = (ax); \
              ((VERTEX_REC *)(aVP))->coords.y = (ay)
#define vertex_data(v, aid, ax, ay)                    \
              *((long *)(aid)) = ((VERTEX_REC *)(v))->id;      \
              *((long *)(ax))  = ((VERTEX_REC *)(v))->coords.x;\
              *((long *)(ay))  = ((VERTEX_REC *)(v))->coords.y

#endif /* FAST_ACCESS */

#endif
