/*
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/

#include "targe2cf.h"
#include "ttypes.h"
#include "tops.h"
#include "cells.h"
#include "vertex.h"
#include "curve.h"
#include "loops.h"
#include "subregs.h"
#include "edges.h"
#include "mctlp.h"
#include "triangle.h"
#include "utils.h"

extern OPTS_REC options; /* import global options */

extern HASH_DATA_REC hash; /* import hash data */

extern LIST front_lists[]; /* import front lists */

CELLS_REC  cells;

#ifdef X_XFACE

#include "xiface.h"

#endif
#ifdef DOS_GXFACE

#include "views.h"
#include "scr.h"
#include "status.h"

#endif



int 
init_cells()
{
  int i;
  
  cells.cell_size = 0.;	         /* 1. */
  cells.sgstd_cells_to_cdim = 0.;/* 2. */
  cells.sgstd_remap_ratio = 0.;	 /* 3. */
  cells.x_origin = 0.;	         /* 4. */
  cells.y_origin = 0.;	         /* 5. */
  cells.cell_of_pull = 0;        /* 6. */
  for (i=0; i<10; i++) 
    cells.neighbors[i] = 0;      /* 7. */
  cells.curr_neighbor = 0;	 /* 8. */
  cells.hyper_array_ptr = NULL;	 /* 9. */
  cells.hyper_n_rows = 0;	 /* 10. */
  cells.hyper_n_cols = 0;	 /* 11. */
  cells.leaf_array_dim = 0.;	 /* 12. */
  cells.n_rows = 0;	         /* 13. */
  cells.n_cols = 0;	         /* 14. */
  cells.iterator_list = NULL;	 /* 15. */
  cells.curr_iter_edge_p = NULL; /* 16. */
  cells.old_hyperar_elems = 0;
  cells.old_leafar_elems = 0;
  return YES;
}




void 
construct_front()
{
  int i;
  
  for (i = 0; i < MAX_FRONT_LISTS; i++)
    front_lists[i] = NULL;
  dummy_cells();
  create_iter_lists();
}



void 
destruct_front()
{
  int i;
  
  for (i = 0; i < MAX_FRONT_LISTS; i++) {
    if (front_lists[i] != NULL) {
      free_list(front_lists[i], NOT_NODES);
      front_lists[i] = NULL;
    }
  }
}



void 
construct_cells(C_REFS_LIST ext_loop)
{
  float xmin, ymin, xmax, ymax;

#define  LARGENUMBER 1.e30 

  cells.sgstd_cells_to_cdim = options.cell_s_2_c_dim;
  cells.sgstd_remap_ratio   = options.remap_ratio;
  xmin=LARGENUMBER; xmax=-LARGENUMBER; ymin=LARGENUMBER; ymax=-LARGENUMBER;
  scan_loop_extent(ext_loop, &xmin, &ymin, &xmax, &ymax);
  setup_cells(xmin, ymin, xmax, ymax);
  remap_to_new_cells();
}



void 
scan_loop_extent(C_REFS_LIST ext_loop,
		      float *xmin, float *ymin, float *xmax, float *ymax)
{
  CURVE_REF_REC *c_r;
  
  c_r = (CURVE_REF_REC *)get_list_next(ext_loop, NULL);
  while (c_r != NULL) {
    update_by_curve_extent(c_r->curve_p, xmin, ymin, xmax, ymax);
    c_r = (CURVE_REF_REC *)get_list_next(ext_loop, c_r);
  }
}



void 
destruct_cells()
{
 /* doing nothing */
}



CELL_ID 
get_which_cell(float at_x, float at_y)
{
  long row, col;
  
  row = (long int)((at_y - cells.y_origin)/cells.cell_size) + 1;
  col = (long int)((at_x - cells.x_origin)/cells.cell_size) + 1;
  return (CELL_ID)( row - 1 ) * cells.n_cols + col;
}




void 
setup_cells(float xmin, float ymin, float xmax, float ymax)
{
  float minmeshsize, maxmeshsize;

  get_maxmin_mesh_cdim(xmin, ymin, xmax, ymax, &minmeshsize, &maxmeshsize);
  cells.cell_size = cells.sgstd_cells_to_cdim * minmeshsize *
    options.magnific_factor;
  assert(cells.cell_size > 0);
  cells.x_origin  = (double)xmin - (double)cells.cell_size * 0.05;
  cells.y_origin  = (double)ymin - (double)cells.cell_size * 0.05;
  cells.n_rows    = (ymax - cells.y_origin) / cells.cell_size + 1;
  cells.n_cols    = (xmax - cells.x_origin) / cells.cell_size + 1;

  if ((double)cells.n_rows*(double)cells.n_cols > (double)MAX_CELLS)
    manip_error("Too large number of cells in the cell grid.");

  hyperarray_setup();
#ifdef X_XFACE
  if (options.X_xface) {
    float tmp1, tmp2;

    tmp1 = 1/maxmeshsize;
    tmp2 = 1/minmeshsize;
    ColorSetupFringeTableByMinMax(EMGetAssocFringeTable(ESIModel ()),
				  tmp1 < 0.75*tmp2? tmp1: tmp2*0.75, tmp2);
  }
#endif  
#ifdef DOS_GXFACE
  if (options.DOS_gxface) {
/* RAW: The color table must be set up here. Also, this is a good place */
/* to map the domain extent to the device coordinates. The input args are */
/* the minimum and maximum coordinates of the subregion to be triangulated. */
	updatelogview(main_view,xmin < 0 ? 1.1*xmin : 0.9*xmin,
				ymin < 0 ? 1.1*ymin : 0.9*ymin,
				xmax < 0 ? 0.9*xmax : 1.1*xmax,
				ymax < 0 ? 0.9*ymax : 1.1*ymax);
	rememberMaxCoords (xmin < 0 ? 1.1*xmin : 0.9*xmin,
			   ymin < 0 ? 1.1*ymin : 0.9*ymin,
			   xmax < 0 ? 0.9*xmax : 1.1*xmax,
			   ymax < 0 ? 0.9*ymax : 1.1*ymax);
	openview (main_view);
  }
#endif
  init_hash_by_length(minmeshsize, maxmeshsize);
}




void 
hyperarray_setup()
{
  long dim1, dim2;
  double tmp1, tmp2;
  long tmp3;

#define TRUNC(N) ((long)(N))

  cells.old_hyperar_elems = cells.hyper_n_rows * cells.hyper_n_cols;
  cells.old_leafar_elems  = cells.leaf_array_dim * cells.leaf_array_dim;
  dim1 = TRUNC(sqrt( (double)cells.n_rows )) + 1;
  dim2 = TRUNC(sqrt( (double)cells.n_cols )) + 1;
  tmp1 = pow((double)LEAF_ARRAY_TOTAL,(double)0.25);
  tmp2 = (double)(min(dim1,dim2));
  tmp3 = max(TRUNC(tmp2/tmp1), TRUNC(MIN_LEAF_DIM));
  cells.leaf_array_dim = TRUNC(tmp3);
  cells.hyper_n_cols   = TRUNC(cells.n_cols/cells.leaf_array_dim) + 1;
  cells.hyper_n_rows   = TRUNC(cells.n_rows/cells.leaf_array_dim) + 1;
  cells.n_cols         = cells.hyper_n_cols * cells.leaf_array_dim;
  cells.n_rows         = cells.hyper_n_rows * cells.leaf_array_dim;
}




void 
dummy_cells()
{
  cells.cell_size = 2.e+30;
  cells.x_origin  = -0.5 * cells.cell_size;
  cells.y_origin  = -0.5 * cells.cell_size;
  cells.n_rows    = 1;
  cells.n_cols    = 1;

  hyperarray_setup();
}



void 
rearrange_cells( float old_cell_size, float new_cell_size )
{
  float xmax, ymax, x, y;

  xmax            = cells.x_origin + cells.n_cols * cells.cell_size;
  ymax            = cells.y_origin + cells.n_rows * cells.cell_size;
  cells.cell_size = new_cell_size;
  assert(cells.cell_size > 0);
  
  cells.n_rows    = 0;
  y               = cells.y_origin;
  do {
      y += cells.cell_size;
      cells.n_rows++;
    } while( y < ymax );
  
  cells.n_cols    = 0;
  x               = cells.x_origin;
  do {
      x += cells.cell_size;
      cells.n_cols++;
    } while( x < xmax );

  hyperarray_setup();
}



void 
remap_to_new_cells()
{
  int fl;
  LIST flptr;
  EDGE_REC *ep;
  VERTEX_REC *v1p = NULL, *v2p = NULL;
  float fx, fy;
  CELL_ID cell;
  sense s;
DBG(
    int print_edge_data(NODE ptr, EDGE_REC *e_p); 
)

  if (!destroy_iter_lists())
    manip_error("Failed destroy_iter_lists");
  if (!create_iter_lists())
    manip_error("Failed create_iter_lists");

  for (fl = 0; fl <= hash.fl_total; fl++) {
    if ((flptr = front_lists[fl]) != NULL) {
      ep = (EDGE_REC *)get_list_next(flptr, NULL);
      while(ep != NULL) {
	s = get_node_val_current(flptr);
	get_edge_vertex_p(ep, &v1p, &v2p);
	get_vertex_coords(v1p, &fx, &fy);
	cell = get_which_cell(fx, fy);
	set_edge_cell(ep, cell);
	if ((EDGE_REC *)add_to_tail_val(get_iterator_list(cell), ep, s) != ep)
	  manip_error("remap_to_new_cells: failed to add to iterator list");
	ep = (EDGE_REC *)get_list_next(flptr, ep);
      }
    }
  }
}
