/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/
 
 
#include "targe2cf.h"
#include <stdio.h>
#ifdef UNIX
#include <unistd.h>
#endif
#include <string.h>
#ifdef MSDOS
#include <io.h>
/*#include <alloc.h>*/
#endif

#include "ttypes.h"
#include "input.h"
#include "cmds.h"
#include "vertex.h"
#include "edges.h"
#include "curve.h"
#include "loops.h"
#include "subregs.h"
#include "mctlp.h"
#include "utils.h"

extern OPTS_REC options; /* this structure holds all options */


static int first_command = 1;

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
def_curve(void)
{
  char *type;
  int id, pos, tot;
  CURVE_REC *c;

  if (first_command && options.qtree)
    bad_syntax("Input file must start by the REGEXT command, when using -q");
  first_command = 0;

  tot = get_tokens_total();
  id   = token_int_val(2);
  type = token_string_val(3);
  if (compare_token_with_keyword(type, "LINE") == YES) {
    add_curve_line(id, token_float_val(4), token_float_val(5),
		   token_float_val(6), token_float_val(7));
    return;
  }
  if (compare_token_with_keyword(type, "DISCRETIZED") == YES) {
    if (IS_EVEN(tot))
      bad_syntax("Missing coordinates!?");
    c = add_curve_discretized(id, token_float_val(4), token_float_val(5),
			      token_float_val(tot-1), token_float_val(tot));
    generate_edges_of_discretized_curve(c);
    return;
  }
  if (compare_token_with_keyword(type, "CIRCLE") == YES) {
    float cx, cy, fx, fy, r = 0.0;
    
    if ((pos = is_keyword_in_line(4 , "CENTER")) > 0) {
      cx = token_float_val(pos + 1);
      cy = token_float_val(pos + 2);
    }
    if ((pos = is_keyword_in_line(4 , "RAD")) > 0) {
      r  = token_float_val(pos + 1);
    }
    if ((pos = is_keyword_in_line(4 , "FIRST")) > 0) {
      fx = token_float_val(pos + 1);
      fy = token_float_val(pos + 2);
    } else {
      fx = cx + r;
      fy = cy;
    }
    add_curve_circle(id, fx, fy, cx, cy);
    return;
  }
  if (compare_token_with_keyword(type, "ARC") == YES) {
    float cx, cy, fx, fy, lx, ly, ix=0, iy=0;
    BOOLEAN have_intermed = NO, opposite_dir = NO;

    fx = token_float_val(4);
    fy = token_float_val(5);
    lx = token_float_val(6);
    ly = token_float_val(7);
    if ((pos = is_keyword_in_line(4 , "CENTER")) > 0) {
      cx = token_float_val(pos + 1);
      cy = token_float_val(pos + 2);
    }
    if ((pos = is_keyword_in_line(4 , "INTERMEDIATE")) > 0) {
      have_intermed = YES;
      ix = token_float_val(pos + 1);
      iy = token_float_val(pos + 2);
    } 
    if ((pos = is_keyword_in_line(4 , "OPPOSITE")) > 0) 
      opposite_dir = YES;
    if ((pos = is_keyword_in_line(4 , "REVERSE")) > 0) 
      opposite_dir = YES;
    if ((pos = is_keyword_in_line(4 , "REV")) > 0) 
      opposite_dir = YES;
    
    add_curve_circ_arc(id, fx, fy, lx, ly, have_intermed, ix, iy, cx, cy,
		       opposite_dir);
    return;
  }
}


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
def_subregion(void)
{
  int pid, mid, sid, i, tot, pos;
  int cid;
  token_type type;
  C_REFS_LIST      ext_loop;
  C_REFS_LIST_LIST int_loops;
  char group[] = "UNKNOWN_GR";
  float thickness;

  if (first_command && options.qtree)
    bad_syntax("Input file must start by the REGEXT command, when using -q");
  first_command = 0;
  
  tot = get_tokens_total();
  
  if ((pos = is_keyword_in_line(3, "PROPERTY")) > 0)
    pid = token_int_val(pos + 1);
  else
    bad_syntax("Property id missing");
  
  mid = 0;
  if ((pos = is_keyword_in_line(3, "MATERIAL")) > 0)
    mid = token_int_val(pos + 1);
  
  sid = 0;
  if ((pos = is_keyword_in_line(3, "SUBSOIL")) > 0)
    sid = token_int_val(pos + 1);
  
  thickness = 0;
  if ((pos = is_keyword_in_line(3, "THICKNESS")) > 0)
    thickness = token_float_val(pos + 1);

  strcpy(group, "UNKNOWN_GR");
  if ((pos = is_keyword_in_line(3, "GROUP")) > 0)
    strcpy(group, get_token_as_string(pos + 1));
  
  if ((pos = is_keyword_in_line(3, "BOUNDARY")) > 0) {
    ext_loop = add_loop();
    for(i = pos + 1; i <= tot; i++) {
      type = get_token_type(i);
      if (type == NUMBER) {
	cid = token_int_val(i);
	if (cid != 0)		/* skip curves of zero id's */
	  append_curve_ref_to_loop(ext_loop, cid);
      } else
	break; /* done for the external boundary */
    }
  } else
    bad_syntax("Boundary keyword missing");
  {
    int hole_keyw_pos;
    C_REFS_LIST  c_r_l;
    static int i;
    
    hole_keyw_pos = is_keyword_in_line(2, "HOLE");
    if (hole_keyw_pos > 0)
      int_loops = make_list();
    else
      int_loops = NULL;
    while(hole_keyw_pos > 0) {
      c_r_l = add_loop();
      add_to_tail(int_loops, c_r_l);
      for(i = hole_keyw_pos + 1; i <= tot; i++) {
	type = get_token_type(i);
	  if (type == NUMBER) {
	    cid = token_int_val(i);
	    if (cid != 0)		/* skip curves of zero id's */
	      append_curve_ref_to_loop(c_r_l, cid);
	  } else {
	    break;
	  }
      }
      hole_keyw_pos = (i <= tot? is_keyword_in_line(i-1, "HOLE"): 0);
    }
  }
  add_subreg(token_int_val(2), pid, mid, sid, thickness, group,
	     ext_loop, int_loops);
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
def_mctl_point(void) 
{
  int pos, id;
  float x, y;
  float cdim, infl_radius;

  if (first_command && options.qtree)
    bad_syntax("Input file must start by the REGEXT command, when using -q");
  first_command = 0;
  
#ifdef INPUT_COMPATIBILITY
  if ((pos = is_keyword_in_line(2, "CONST")) > 0 ||
      (pos = is_keyword_in_line(2, "CONSTANT")) > 0)
    add_const_cdim(token_float_val(pos + 1));
  else if ((pos = is_keyword_in_line(2, "TRIANGLE")) > 0 ||
	   (pos = is_keyword_in_line(2, "TRI")) > 0) {
    float x1, y1, x2, y2, x3, y3, v1, v2, v3;

    pos = is_keyword_in_line(pos, "XY");
    x1 = token_float_val(pos+1);
    y1 = token_float_val(pos+2);
    x2 = token_float_val(pos+3);
    y2 = token_float_val(pos+4);
    x3 = token_float_val(pos+5);
    y3 = token_float_val(pos+6);
    pos = is_keyword_in_line(pos, "CDIM");
    v1 = token_float_val(pos+1);
    v2 = token_float_val(pos+2);
    v3 = token_float_val(pos+3);
    add_mctl_triangle(x1, y1, x2, y2, x3, y3, v1, v2, v3,
                      1, 1, 1, 0, 0, 0);
  } else {
    x = y = cdim = infl_radius = 0.;
    
    id = token_int_val(2);
    if ((pos = is_keyword_in_line(3, "XY")) > 0) {
      x = token_float_val(pos + 1);
      y = token_float_val(pos + 2);
    }
    if ((pos = is_keyword_in_line(3, "NEAR")) > 0) {
      cdim = token_float_val(pos + 1);
    }
    if ((pos = is_keyword_in_line(3, "INFLUENCE")) > 0) {
      infl_radius = token_float_val(pos + 1);
    }
    if (cdim == 0. || infl_radius == 0.)
      bad_syntax("Mesh-size and influence radius must be non-zero.");
    
    add_mctl_point(id, x, y, cdim, infl_radius);
  }
#else
  manip_error("Cannot handle M-CTL-POINT command");
#endif
}


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
def_fixed_point(void)
{
  int id, pos, curve_id;
  float x, y;

  if (first_command && options.qtree)
    bad_syntax("Input file must start by the REGEXT command, when using -q");
  first_command = 0;
  
  if (get_token_type(2) == NUMBER)
    id = token_int_val(2);
  else
    bad_syntax("Fixed-point id missing!?");
  
  if ((pos = is_keyword_in_line(2, "XY")) > 0) {
    x = token_float_val(pos + 1);
    y = token_float_val(pos + 2);
  } else
    bad_syntax("Fixed-point id coordinates missing!?");
  
  curve_id = 0;
  if ((pos = is_keyword_in_line(2, "CURVE")) > 0) 
    curve_id = token_int_val(pos + 1);
                                       /* DUMMY DATA */
  add_fixed_point(add_vertex(id, x, y,     1, 1, 0),       curve_id);
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
print_cmd(void)
{
  int tot, i;
  token_type type;
  double dval;
  long lval;

  if (first_command && options.qtree)
    bad_syntax("Input file must start by the REGEXT command, when using -q");
  first_command = 0;
  
  if (!options.silent_as_a_tomb) {
    tot = get_tokens_total();
    
    for (i=2; i <= tot; i++)    {
      type = get_token_type(i);
      switch((int)type)	{
      case (int)NUMBER:
	dval = get_token_as_double(i);
	lval = (long)dval;
	if ((double)lval == dval) {
	  if (lval > 32000 || -lval > 32000)
	    printf(" %ld", (long)lval);
	  else
	    printf(" %d", (int)lval);
	} else
	  printf(" %f", dval);
	break;
      case (int)STRING:
	printf(" %s", get_token_as_string(i));
	break;
      default:
	printf("\n  %d type = UNKNOWN",i);
      }
    }
    printf("\n");
  }
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
def_constant_mesh_size()	/* MC-CONST SIZE */
{
  if (first_command && options.qtree)
    bad_syntax("Input file must start by the REGEXT command, when using -q");
  first_command = 0;
  
  add_const_cdim(token_float_val(2));
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
def_plain_ctl_point()	/* MC-PLAIN X Y SIZE INFLUENCE_RADIUS */
{
  if (first_command && options.qtree)
    bad_syntax("Input file must start by the REGEXT command, when using -q");
  first_command = 0;
  
  add_mctl_point(token_int_val(2), token_float_val(3), token_float_val(4),
		 token_float_val(5), token_float_val(6));
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
def_by_triangle_control()	/* MC-TRI X1 Y1 X2 Y2 X3 Y3 S1 S2 S3 */
{
  if (first_command && options.qtree)
    bad_syntax("Input file must start by the REGEXT command, when using -q");
  first_command = 0;
  
#ifdef ANISOTROPIC_MESH
  if (options.anisotropic_mesh) 
    add_mctl_triangle(token_float_val(2), token_float_val(3),
                      token_float_val(4), token_float_val(5),
                      token_float_val(6), token_float_val(7),
                      token_float_val(8), token_float_val(9),
                      token_float_val(10), token_float_val(11), 
                      token_float_val(12), token_float_val(13), 
                      token_float_val(14), token_float_val(15), 
                      token_float_val(16));
  else
#endif
    add_mctl_triangle(token_float_val(2), token_float_val(3),
                      token_float_val(4), token_float_val(5),
                      token_float_val(6), token_float_val(7),
                      token_float_val(8), token_float_val(9),
                      token_float_val(10),
                      1, 1, 1, 0, 0, 0);
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void def_by_triangle_file()	/* MC-FILE file */
{
  int numlines, nread, i;
  float a[15] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
  FILE *stream;

  if (first_command && options.qtree)
    bad_syntax("Input file must start by the REGEXT command, when using -q");
  first_command = 0;

  stream = fopen(token_string_val(2), "r");

  if (stream == (FILE *)NULL)
    manip_error("Triangle stream could not be opened");
  else {
    if ((nread = fread(&numlines, sizeof(int), 1, stream)) != 1)
      manip_error("Error reading triangle stream");
    for (i = 0; i < numlines; i++) {
      if (options.anisotropic_mesh) {
        if ((nread = fread(a, sizeof(float), 15, stream)) != 15)
          manip_error("Error reading triangle stream");
      } else {
        if ((nread = fread(a, sizeof(float), 9, stream)) != 9)
          manip_error("Error reading triangle stream");
      }
      add_mctl_triangle(a[0], a[1], a[2], a[3], a[4], a[5], a[6],
                        a[7], a[8], a[9], a[10], a[11], a[12], a[13], a[14]);
    }
    fclose(stream);
  }
}


void
set_region_extent()
{
  float minx, miny, maxx, maxy;

  first_command = 0;
  
  minx = get_token_as_double(2);
  miny = get_token_as_double(3);
  maxx = get_token_as_double(4);
  maxy = get_token_as_double(5);
  set_subregions_bbox(minx, miny, maxx, maxy);
}
