/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/

#include "Esimple.h"


static void 
LoadMesh(Widget w, XtPointer text_ptr, XtPointer call_data);


void
ESICustomize(Widget parent_pane)
{
  Widget load_mesh;

  load_mesh = ESIAddButton("load_mesh", "Load mesh", commandWidgetClass,
                           parent_pane, NULL, 0, LoadMesh, NULL);
}

static void
clean_up_old_mesh(void);
static void
read_new_mesh(void);
static void
read_mesh(char *geom_file_name);


static void 
LoadMesh(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  char *name;
  FILE *fp;
  
  fp = ESISelFile("Select *.out file", "OK", "CANCEL", NULL, ".", "r", &name);
  if (fp != NULL) {
    fclose(fp);
    read_mesh(name);
  }
}


typedef struct node_rec {
  double x, y;
}              node_rec;



static long num_nodes = 0;
static long num_triangles = 0;
static node_rec *node_array = NULL;
static FILE *handle = NULL;
static EModel *elixir_model = NULL;
static EPixel tri_color = 0;

static void
read_mesh(char *geom_file_name)
{
  BOOLEAN status;
  elixir_model = ESIModel();
  
  tri_color = ColorGetPixelFromString("red", &status);
  if (!status) tri_color = 0;
  EASValsSetColor(tri_color);
  EASValsSetShrink(0.85);
  EASValsSetFillStyle(FILL_SOLID);

  clean_up_old_mesh();

  handle = fopen(geom_file_name,"r");
  read_new_mesh();
  fclose(handle);
}



static void
clean_up_old_mesh(void)
{
  free(node_array);
  node_array = NULL;
  elixir_model = ESIModel();
  EMDestroyAllGraphics(elixir_model);
}


#define LINE_LENGTH 256
static char line[LINE_LENGTH];
#define GET_LINE() fgets(line, LINE_LENGTH-1, handle)

static void
read_new_mesh(void)
{
  long i, nn, nt;
  float nx, ny;
  GraphicObj *g_p;
  long n1, n2, n3;
  WCRec pnts[3];
  
  GET_LINE();                   /* file name */
  GET_LINE();                   /* number of nodes, number of triangles */
  sscanf(line,"%d %d", &num_nodes, &num_triangles);
  GET_LINE();                   /* *NODES */

  node_array = calloc(num_nodes+1, sizeof(node_rec));
  
  for(i = 0; i < num_nodes; i++) {
    GET_LINE();
    sscanf(line, "%d %f %f", &nn, &nx, &ny);
    node_array[nn].x = nx;
    node_array[nn].y = ny;
  }

  GET_LINE();                   /* *TR3N */
  
  for(i = 0; i < num_triangles; i++) {
    GET_LINE();
    sscanf(line, "%d %d %d %d", &nt, &n1, &n2, &n3);
    pnts[0].x = node_array[n1].x;
    pnts[0].y = node_array[n1].y;
    pnts[0].z = 0;
    pnts[1].x = node_array[n2].x;
    pnts[1].y = node_array[n2].y;
    pnts[1].z = 0;
    pnts[2].x = node_array[n3].x;
    pnts[2].y = node_array[n3].y;
    pnts[2].z = 0;
    g_p = CreateTriangle3D(pnts);
    EGWithMaskChangeAttributes(COLOR_MASK| SHRINK_MASK | FILL_MASK, g_p);
    EMNoDrawAddGraphicsToModel(elixir_model, g_p);
  }
}

int 
main(int argc, char **argv)
{
  ESIBuildInterface(ESI_GRAPHIC_EDITOR_MASK, argc, argv);
  ESIPopup();
  ESIHandleCmd("COLOR yellow");
  ESIHandleCmd("SHRINK 0.9");
  ESIHandleCmd("NEWFRAME"); 
  if (argc > 1) {
    int i = 1;
    char buf[1024];
    while (i < argc) {
      read_mesh (argv[i]);
      sprintf (buf, "Mesh from %s", argv[i]);
      ESIEventLoop (TRUE, buf);
      i++;
    }
  }
  ESIPopupAndRun();
}
