/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/

#include "targe2cf.h"
#include "ttypes.h"

#include "tops.h"
#include "curve.h"
#include "loops.h"
#include "utils.h"

LOOPS_LIST loops_list;

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 /* make_loops_list  */

void make_loops_list()
{
  loops_list = make_list();
  assert(loops_list != NULL);
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 /* add_loop  */

LIST add_loop()
{
  LOOP_REC *l_r;

  make_and_add_to_tail_val(loops_list, l_r, LOOP_REC, "add_loop", 0);
  l_r->curve_ref_list = make_list();
  assert(l_r->curve_ref_list != NULL);
  return l_r->curve_ref_list;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void append_curve_ref_to_loop(LIST loop, int signed_curve_id)
{
  CURVE_REF_REC *c_r;
  CURVE_REC     *c;

  assert(loop != NULL);
  c = find_curve_by_id(abs(signed_curve_id));
  if (c == NULL)    {
    char buffer[80];
    sprintf(buffer, "No such curve id: %d", abs(signed_curve_id));
    manip_error(buffer);
  }
  make_and_add_to_tail_val(loop, c_r, CURVE_REF_REC,
			   "append_curve_ref_to_loop",
			   signed_curve_id);
  c_r->curve_sense = signed_curve_id > 0? AS_DEFINED_SENSE: REVERSED_SENSE;
  c_r->curve_p     = c;
}


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

int mesh_loop(struct _SUBREG_REC *s_r, C_REFS_LIST c_r_l)
{
  assert(c_r_l != NULL);
  for_all_list((LIST)c_r_l, NULL, mesh_curve);
  return YES;
}
