/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/

#include "targe2cf.h"
#include "ttypes.h"

#include <stdio.h>
#ifdef UNIX
#include <unistd.h>
#endif
#include <string.h>
#include <time.h>

#include "targe2.h"
#include "tops.h"
#include "log.h"

#include "subregs.h"
#include "vertex.h"
#include "edges.h"
#include "loops.h"
#include "mctlp.h"
#include "qtree.h"
#include "mesh.h"
#include "triangle.h"
#include "cells.h"



#ifdef X_XFACE
#include "xiface.h"
#endif

#ifdef DOS_GXFACE

#include <graphics.h>
#include "views.h"
#include "scr.h"
#include "status.h"

#endif

extern OPTS_REC options; /* import global options */

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void
remove_3_and_4_connectivity_vertices(void);

void 
mesh_region()
{
  overall_startup();
  mesh_subregions();
  if (options.quality_enhancements)
    remove_3_and_4_connectivity_vertices();
  
  if (options.smooth_final_mesh)
    smooth_mesh();
  overall_cleanup();
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void overall_startup()
{
#ifdef Q_TREE
  qt_prepare_for_use();
#endif
  init_cells();
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void overall_cleanup()
{
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void mesh_interior(SUBREG_REC *s_r)
{
#ifdef X_XFACE
  if (options.X_xface)
    ERptStatusMessage(1, 1, "Inside triangulation loop");
#endif
#ifdef DOS_GXFACE
  if (options.DOS_gxface) {
/* RAW: message */
    Print_status(1);
  }
#endif  
  while(pull_edge_from_front() == YES) {
#ifdef X_XFACE
    if (options.X_xface) {
      xiface_loop(options.stop_for_each, "Triangulating: ");
    }
#endif
#ifdef DOS_GXFACE
    if (options.DOS_gxface) {
/* RAW: message -- watch for interrupt requests and handle interrupts */
	Print_status(1);
    }
#endif
    construct_triangle(s_r);
  }
#ifdef X_XFACE
  if (options.X_xface) {
    xiface_loop(YES, "Subregion triangulated.");
  }
#endif
#ifdef DOS_GXFACE
  if (options.DOS_gxface) {
/* RAW: message */
	Print_status(1);
  }
#endif 
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */


void 
smooth_mesh(void)
{
  int pass;

  for (pass = 1; pass <= MAX_SMOOTH_PASSES; pass++)
    smooth_pass(options.smooth_final_mesh);
#ifdef X_XFACE
  if (options.X_xface) {
#ifdef Q_TREE
    DBG( undraw_qtree(); )
#endif
    draw_edges();

    draw_triangles();
    xiface_loop(YES, "Smoothed.");
  }
#endif
#ifdef DOS_GXFACE
  if (options.DOS_gxface) {
    clearviewport();
    setcolor(FRAME);
    frameview (main_view->xlogmin,main_view->ylogmin,
	       main_view->xlogmax,main_view->ylogmax);
    setcolor(DRAW_AREA_INK);
    
    draw_edges();
    draw_triangles();
    while (!kbhit());
    while(!Check_status(0));
/* RAW: message */
  }
#endif
}

