
/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/

#include "targe2cf.h"
#include "ttypes.h"
#include "targe2.h"
#include "tops.h"
#include "log.h"

#include "subregs.h"
#include "vertex.h"
#include "edges.h"
#include "loops.h"
#include "mctlp.h"
#include "mesh.h"
#include "triangle.h"



extern PULLED_EDGE_REC pulled_edge; /* import pulled-edge */

static double shared_point_eps = 0;

static VERTEX_REC *
can_use_intersected_edge(EDGE_REC_P intersected);
static void 
small_angle_check(VERTEX_REC **selvertexp,
                  EDGE_REC **edge1_p, EDGE_REC **edge2_p,
                  float *selvx, float *selvy);
static BOOLEAN 
last_resort(VERTEX_REC **selvertexp, float *selvx, float *selvy,
            EDGE_REC **edge1_p, EDGE_REC **edge2_p,
            EDGE_REC **isectswith_pp);

#define IPOL_LINE(eta, v1, v2) (((1. - (eta)) * (v1)) + ((eta) * (v2)))
			
#define PE(aFIELD) pulled_edge.aFIELD
#define IPX PE(ideal_point.x)
#define IPY PE(ideal_point.y)
#define FX  PE(first.x)
#define FY  PE(first.y)
#define LX  PE(last.x)
#define LY  PE(last.y)

#define AVRG(theV1, theV2) (0.5*((theV1)+(theV2)))

#define SUCCESS YES
#define FAILURE NO

extern OPTS_REC options; /* import global options */

#define MAX_SVRTS  6

static struct {
  int       cur;
  float     largest_dist;
  int       largest_dist_at;
  float     min_dist;
  int       min_dist_at;
  struct {
    float       x;
    float       y;
    VERTEX_REC *p;
    float       d;
  } svrts[MAX_SVRTS];
} svrts_buf;

#define CLEAR_SVRTS_BUF                                                     \
   {                                                                        \
      svrts_buf.largest_dist    = 0;                                        \
      svrts_buf.largest_dist_at = -1;                                       \
      svrts_buf.min_dist        = 0;                                        \
      svrts_buf.min_dist_at     = -1;                                       \
      for (svrts_buf.cur = 0; svrts_buf.cur < MAX_SVRTS; svrts_buf.cur++) { \
        svrts_buf.svrts[svrts_buf.cur].x = 0;                               \
        svrts_buf.svrts[svrts_buf.cur].y = 0;                               \
        svrts_buf.svrts[svrts_buf.cur].p = NULL;                            \
        svrts_buf.svrts[svrts_buf.cur].d = 0;                               \
      }                                                                     \
    } 

#define SVRTS_BUF_PUT(aX, aY, aDIST, aVP, aLABEL)                       \
  {                                                                     \
    for (svrts_buf.cur = 0; svrts_buf.cur < MAX_SVRTS; svrts_buf.cur++){\
      if (svrts_buf.svrts[svrts_buf.cur].p == NULL) {                   \
        svrts_buf.svrts[svrts_buf.cur].x = aX;                          \
        svrts_buf.svrts[svrts_buf.cur].y = aY;                          \
        svrts_buf.svrts[svrts_buf.cur].p = aVP;                         \
        svrts_buf.svrts[svrts_buf.cur].d = aDIST;                       \
        if (svrts_buf.largest_dist_at == -1) {                          \
          svrts_buf.largest_dist_at = svrts_buf.cur;                    \
          svrts_buf.largest_dist    = aDIST;                            \
        }                                                               \
        goto aLABEL;                                                    \
      } else {                                                          \
        if (svrts_buf.svrts[svrts_buf.cur].p == aVP) goto aLABEL;       \
        if (svrts_buf.svrts[svrts_buf.cur].d > svrts_buf.largest_dist){ \
          svrts_buf.largest_dist_at = svrts_buf.cur;                    \
          svrts_buf.largest_dist    = svrts_buf.svrts[svrts_buf.cur].d; \
        }                                                               \
      }                                                                 \
    }                                                                   \
    if (   svrts_buf.largest_dist_at >= 0                               \
        && svrts_buf.largest_dist_at <  MAX_SVRTS) {                    \
      if (aDIST < svrts_buf.largest_dist) {                             \
        svrts_buf.cur = svrts_buf.largest_dist_at;                      \
        svrts_buf.svrts[svrts_buf.cur].x = aX;                          \
        svrts_buf.svrts[svrts_buf.cur].y = aY;                          \
        svrts_buf.svrts[svrts_buf.cur].p = aVP;                         \
        svrts_buf.svrts[svrts_buf.cur].d = aDIST;                       \
        goto aLABEL;                                                    \
      }                                                                 \
    } else {                                                            \
      manip_error("no space in svrts_buf?");                            \
    }                                                                   \
  aLABEL: ;                                                              \
  }      

#define SVRTS_BUF_GET(aX, aY, aVP)                                        \
  {                                                                       \
    aX = aY = 0; aVP = NULL;                                              \
    svrts_buf.min_dist_at = -1;                                           \
    svrts_buf.min_dist    = FLT_MAX;                                      \
    for (svrts_buf.cur = 0; svrts_buf.cur < MAX_SVRTS; svrts_buf.cur++) { \
      if (svrts_buf.svrts[svrts_buf.cur].p != NULL)                       \
        if (svrts_buf.svrts[svrts_buf.cur].d < svrts_buf.min_dist) {      \
           svrts_buf.min_dist_at = svrts_buf.cur;                         \
           svrts_buf.min_dist    = svrts_buf.svrts[svrts_buf.cur].d;      \
         }                                                                \
    }                                                                     \
    if (svrts_buf.min_dist_at    >= 0                                     \
        && svrts_buf.min_dist_at <  MAX_SVRTS) {                          \
        svrts_buf.cur = svrts_buf.min_dist_at;                            \
        aX = svrts_buf.svrts[svrts_buf.cur].x;                            \
        svrts_buf.svrts[svrts_buf.cur].x = 0;                             \
        aY = svrts_buf.svrts[svrts_buf.cur].y;                            \
        svrts_buf.svrts[svrts_buf.cur].y = 0;                             \
        aVP = svrts_buf.svrts[svrts_buf.cur].p;                           \
        svrts_buf.svrts[svrts_buf.cur].p = NULL;                          \
    }                                                                     \
  }


static int
num_of_flanking_edges(VERTEX_REC *v, VERTEX_REC *v1_p, VERTEX_REC *v2_p);


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
construct_triangle(SUBREG_REC *s_r)
{
  float ecx, ecy, vx, vy, ipechalf, ir, dist;
  EDGE_REC_P ep;
  EDGE_REC_P edge1_p = NULL;
  EDGE_REC_P edge2_p = NULL;
  EDGE_REC_P isectswith_p = NULL;
  sense sense; 
  VERTEX_REC *vert1p = NULL, *vert2p = NULL, *vertp = NULL;
  VERTEX_REC *scndp = NULL, *tmp_vp = NULL;
  VERTEX_REC *svp = NULL;
  float svx, svy;
  float eta, etastep, mineta, minheight;
  int num_flanking;
  float x, y;
  float Pischweitz, pischweitz_fact = 2, a, b, c, psw;
  float mincd, maxcd;
  BOOLEAN fixed_selected;

  get_cdim_extremes(&mincd, &maxcd);
  shared_point_eps = 0.01 * mincd;
#define SCALAR_PROD(vx1, vy1, vx2, vy2) ((vx1)*(vx2)+(vy1)*(vy2))
#define INSIDE_HALF_PLANE (vx * (x - ecx) + vy * (y - ecy) > 0.)
#define NOT_TOO_LARGE_STRETCH 1 /* (line_length(x, y, ecx, ecy) < 4*ipechalf)*/
/* To prevent glueing of vertices of adjacent triangles in a row */
#define MAKING_ACUTE_ANGLES                                                 \
    (   ((LX-FX)*(x-(FX+0.49*(FX-LX))) + (LY-FY)*(y-(FY+0.49*(FY-LY))) > 0) \
     && ((FX-LX)*(x-(LX+0.49*(LX-FX))) + (FY-LY)*(y-(LY+0.49*(LY-FY))) > 0))
/* With forced vertices:  better shapes by limiting further the location */
#define WITHIN_CIRCLE_AROUND_MIDDLE                                  \
    (line_length(x, y, 0.5*(IPX+ecx), 0.5*(IPY+ecy)) <= ipechalf)

  /* Now comes the most beautiful piece of code that does the most strange */
  /* things ever seen. */
  Pischweitz = options.Pischweitz;
  c = ((options.Pischweitz * pischweitz_fact - 1)/0.3125 + 1 -
       options.Pischweitz);
  b = 4 * options.Pischweitz - 4 + 2 * c;
  a = options.Pischweitz - 1 - b - c;
  psw =  PE(mf_at_ideal) * PE(cdim_at_center);
  Pischweitz =
    a * psw * psw * psw * psw + b * psw * psw * psw + c *  psw * psw + 1;
  
  /* This is normal again (I mean no lunacy) */
  ecx      = AVRG(FX, LX); ecy      = AVRG(FY, LY);
  vx       = -(LY - FY);   vy       = LX - FX;
  ipechalf = 0.5 * line_length(ecx, ecy, IPX, IPY);
  ir   = (min(PE(cdim_at_first), PE(cdim_at_last))
          * AVRG(PE(mf_at_first), PE(mf_at_last)));
  fixed_selected = NO;
  select_fixed_point(IPX, IPY, FX, FY, LX, LY, ecx, ecy, ir, &svp, &svx, &svy);
  if (svp != NULL) fixed_selected = YES;
  CLEAR_SVRTS_BUF;
  if (svp != NULL) {
    dist = line_length(IPX, IPY, svx, svy);
    SVRTS_BUF_PUT(svx, svy, dist, svp, saved_fixed);
  } else {
    first_front_edge(&ep, &sense);
    while (ep != NULL) {
      if (ep != PE(e_p)) {
	get_edge_vertex_p(ep, &vert1p, &vert2p);
	vertp = vert1p, scndp = vert2p;
	if (sense == REVERSED_SENSE) 
	  vertp = vert2p, scndp = vert1p;
	if (vertp == PE(v1_p) || vertp == PE(v2_p)) {
	  tmp_vp = vertp, vertp = scndp, scndp = tmp_vp;
	}
	get_vertex_coords(vertp, &x, &y);
	if (INSIDE_HALF_PLANE) {
	  dist = line_length(x, y, IPX, IPY);
#ifdef ANISOTROPIC_MESH
          if (options.anisotropic_mesh) {
            if ((dist < (ideal_edge_length(IPX, IPY, x, y,
                                           PE(cdim_at_ideal),
                                           Pischweitz * PE(mf_at_ideal),
                                           PE(ang_at_ideal))))
                && NOT_TOO_LARGE_STRETCH) {
              SVRTS_BUF_PUT(x, y, dist, vertp, saved_a);
            }
          } else {
#endif
            if (dist < ir && NOT_TOO_LARGE_STRETCH) {
              SVRTS_BUF_PUT(x, y, dist, vertp, saved_b);
            }
#ifdef ANISOTROPIC_MESH
          }
#endif
        }
      }
      next_front_edge(&ep, &sense);
    }
  }
  /* Now we have found vertices in a buffer. */
  SVRTS_BUF_GET(svx, svy, svp);
  x=svx; y=svy;
  while (svp != NULL) {
    isectswith_p = NULL;
    num_flanking = num_of_flanking_edges(svp, PE(v1_p), PE(v2_p));
    if (num_flanking == 2) {
      update_front(svp, svx, svy, &edge1_p, &edge2_p);
      break;
    } else if ((num_flanking == 1
                && MAKING_ACUTE_ANGLES)
               || num_flanking == 0) {
      if (no_intersections(svp, svx, svy, NULL, &isectswith_p)) {
        VERTEX_REC *Svp = NULL;
        float Svx = svx, Svy = svy;
        EDGE_REC *Ep1 = NULL, *Ep2 = NULL;
        
        if (!fixed_selected) small_angle_check(&Svp, &Ep1, &Ep2, &Svx, &Svy); 
        if (Svp == NULL) {
          update_front(svp, svx, svy, &edge1_p, &edge2_p);
        } else {
          svp = Svp; svx = Svx; svy = Svy; edge1_p = Ep1; edge2_p = Ep2;
        }
        break;
      }
    }
    SVRTS_BUF_GET(svx, svy, svp);
  }
  if (svp == NULL) {
    isectswith_p = NULL;
    etastep      = -0.04999;
    mineta       = 0.1;
#define REGULAR_START()                                                  \
 { eta = 1.0 + 0.99999 * fabs(etastep); isectswith_p = NULL; continue; }
    for (eta = 1.25; eta >= mineta; eta += etastep) {
      svx = IPOL_LINE(eta, ecx, IPX);
      svy = IPOL_LINE(eta, ecy, IPY);
      if (no_intersections(svp, svx, svy, NULL, &isectswith_p)) {
        if (eta > 1) { /* first test is done for an over-stretched triangle */
          REGULAR_START();
        }
	small_angle_check(&svp, &edge1_p, &edge2_p, &svx, &svy);
	if (svp == NULL) 
	  goto continue_construct;
	else
	  goto continue_have_it;
      } else {
        if (eta > 1) { /* first test is done for an over-stretched triangle */
          REGULAR_START();
        }
        if ((svp = can_use_intersected_edge(isectswith_p)) != NULL) {
          get_vertex_coords(svp, &svx, &svy);
          update_front(svp, svx, svy, &edge1_p, &edge2_p);
          goto continue_have_it;
        }
        svp = NULL;
      }
    }
    if (last_resort(&svp, &svx, &svy, &edge1_p, &edge2_p, &isectswith_p)) 
      goto continue_have_it;
    else {
      manip_error("Last_resort failed");
      svp = NULL;
      goto continue_have_it;
    }
  continue_construct:
    svp = add_vertex(0, svx, svy, NO_CDIM_AVAILABLE, 1, 0);
    edge1_p = add_edge(PE(v1_p), svp, NULL);
    add_to_front(edge1_p, AS_DEFINED_SENSE, line_length(svx, svy, FX, FY));
    edge2_p = add_edge(svp, PE(v2_p), NULL);
    add_to_front(edge2_p, AS_DEFINED_SENSE, line_length(svx, svy, LX, LY));
  }
 continue_have_it:
  if (svp == NULL) {
    reject_edge(PE(e_p), PE(sense), PE(length));
  } else {
    adj_fixed_points(svp);
    /* Note: the edges are ordered so that to traverse them you */
    /* go counterclockwise when looking against Z axis */
    add_triangle(PE(e_p), edge2_p, edge1_p, s_r);
  }
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

BOOLEAN 
no_intersections(VERTEX_REC *selvertexp, float selvx, float selvy,
                 EDGE_REC_P to_ignore, EDGE_REC_P *isectswith_pp)
{
  float edge1x1, edge1y1, edge1x2, edge1y2;
  float edge2x1, edge2y1, edge2x2, edge2y2;
  float edge1boxlx, edge1boxly, edge1boxux, edge1boxuy;
  float edge2boxlx, edge2boxly, edge2boxux, edge2boxuy;
  float x1, y1, x2, y2, x, y, ac, cos1, cos2;
  BOOLEAN ignore = NO;
  VERTEX_REC *v1_p = NULL, *v2_p = NULL, *vp = NULL;
  EDGE_REC_P ep;
  sense s;
  int n, k;
  EDGE_REC_P er[MAX_CONNECTIVITY];

  edge1x1 = FX;
  edge1y1 = FY;
  edge1x2 = selvx;
  edge1y2 = selvy;
  edge2x1 = selvx;
  edge2y1 = selvy;
  edge2x2 = LX;
  edge2y2 = LY;

#define BOX_LIMITS(x1, y1, x2, y2, boxlx, boxly, boxux, boxuy) \
                    if (x1 < x2) { \
                      boxlx = x1; boxux = x2; \
                    } else {\
                      boxlx = x2; boxux = x1; } \
                    if (y1 < y2) { \
                      boxly = y1; boxuy = y2; \
                    } else { \
                      boxly = y2; boxuy = y1; }
		
  BOX_LIMITS(edge1x1, edge1y1, edge1x2, edge1y2,
	     edge1boxlx, edge1boxly, edge1boxux, edge1boxuy)
  BOX_LIMITS(edge2x1, edge2y1, edge2x2, edge2y2,
	     edge2boxlx, edge2boxly, edge2boxux, edge2boxuy)

  if (*isectswith_pp != NULL) {
    if (*isectswith_pp != to_ignore) {
      get_edge_vertex_p(*isectswith_pp, &v1_p, &v2_p);
      get_vertex_coords(v1_p, &x1, &y1);
      get_vertex_coords(v2_p, &x2, &y2);
      if (edges_intersect(edge1x1, edge1y1, edge1x2, edge1y2,
			  edge1boxlx, edge1boxly, edge1boxux, edge1boxuy,
			  x1, y1, x2, y2)) 
	return NO;
      if (edges_intersect(edge2x1, edge2y1, edge2x2, edge2y2,
			  edge2boxlx, edge2boxly, edge2boxux, edge2boxuy,
			  x1, y1, x2, y2)) 
	return NO;
    }
  }
  first_front_edge(&ep, &s);
  while (ep != NULL) {
    if (ep != to_ignore) { 
      get_edge_vertex_p(ep, &v1_p, &v2_p);
      ignore = NO;
      if (selvertexp != NULL) {
	if ((selvertexp == v1_p) && (v2_p == PE(v1_p) || v2_p == PE(v2_p))) {
	  ignore = YES;
	} else {
	  if ((selvertexp == v2_p) && (v1_p == PE(v1_p) || v1_p == PE(v2_p))) 
	    ignore = YES;
	}
      }
      if (!ignore) {
	get_vertex_coords(v1_p, &x1, &y1);
	get_vertex_coords(v2_p, &x2, &y2);
	if (edges_intersect(edge1x1, edge1y1, edge1x2, edge1y2,
			    edge1boxlx, edge1boxly, edge1boxux, edge1boxuy,
			    x1, y1, x2, y2)) {
	  *isectswith_pp = ep;
	  return NO;
	}
	if (edges_intersect(edge2x1, edge2y1, edge2x2, edge2y2,
			    edge2boxlx, edge2boxly, edge2boxux, edge2boxuy,
			    x1, y1, x2, y2)) {
	  *isectswith_pp = ep;
	  return NO;
	}
      }
    }
    next_front_edge(&ep, &s);
  }
  *isectswith_pp = NULL;

  /* check possible containment of some edges within the proposed triangle */
  cos1 = (  (LX - FX) * (selvx - FX) + (LY - FY) * (selvy - FY));
  cos1 /= (PE(length) * line_length(selvx, selvy, FX, FY));
  cos2 = (  (FX - LX) * (selvx - LX) + (FY - LY) * (selvy - LY));
  cos2 /= (PE(length) * line_length(selvx, selvy, LX, LY));
  get_vert_references(PE(v1_p), &n, er);
  for (k = 0; k < n; k++) {
    if (*(er+k) != PE(e_p)) {
      get_edge_vertex_p(*(er+k), &v1_p, &v2_p);
      vp = NULL;
      if      (v1_p == PE(v1_p) && v2_p != selvertexp) vp = v2_p;
      else if (v2_p == PE(v1_p) && v1_p != selvertexp) vp = v1_p;
      if (vp != NULL) {
        get_vertex_coords(vp, &x, &y);
        if (((FY - LY) * (x - FX) + (LX - FX) * (y - FY)) > 0){/* half-plane */
          ac = ((LX - FX) * (x - FX) + (LY - FY) * (y - FY));
          ac /= (PE(length) * line_length(x, y, FX, FY)); /* angle */
          if (ac >= cos1 || cos1 - ac < 0.0001) return NO;
        }
      }
    }
  }
  get_vert_references(PE(v2_p), &n, er);
  for (k = 0; k < n; k++) {
    if (*(er+k) != PE(e_p)) {
      get_edge_vertex_p(*(er+k), &v1_p, &v2_p);
      vp = NULL;
      if      (v1_p == PE(v2_p) && v2_p != selvertexp) vp = v2_p;
      else if (v2_p == PE(v2_p) && v1_p != selvertexp) vp = v1_p;
      if (vp != NULL) {
        get_vertex_coords(vp, &x, &y);
        if (((FY - LY) * (x - LX) + (LX - FX) * (y - LY)) > 0){/* half-plane */
          ac = ((FX - LX) * (x - LX) + (FY - LY) * (y - LY));
          ac /= (PE(length) * line_length(x, y, LX, LY)); /* angle */
          if (ac >= cos2 || cos2 - ac < 0.0001) return NO;
        }
      }
    }
  }
  return YES;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

BOOLEAN 
edges_intersect(float x1, float y1, float x2, float y2,
                float boxlx, float boxly, float boxux, float boxuy,
                float x1prime, float y1prime,
                float x2prime, float y2prime)
{      
  double delta1, delta2, delta, drev, t, s;
  double x2prime_x1prime, y2prime_y1prime;
  double x2_x1, y2_y1, x1prime_x1, y1prime_y1;
  double id_eps;

/* check bounding boxes */
  if ((min(x1prime, x2prime)) > boxux) 
    return NO;
  if ((min(y1prime, y2prime)) > boxuy) 
    return NO;
  if ((max(x1prime, x2prime)) < boxlx) 
    return  NO;
  if ((max(y1prime, y2prime)) < boxly) 
    return  NO;

/* check for shared points */
  if ((fabs(x1 - x1prime) < shared_point_eps)
      &&
      (fabs(y1 - y1prime) < shared_point_eps))
    return NO; /* shared point */
  if ((fabs(x2 - x1prime) < shared_point_eps)
      &&
      (fabs(y2 - y1prime) < shared_point_eps))
    return NO; /* shared point */
  if ((fabs(x1 - x2prime) < shared_point_eps)
      &&
      (fabs(y1 - y2prime) < shared_point_eps))
    return NO; /* shared point */
  if ((fabs(x2 - x2prime) < shared_point_eps)
      &&
      (fabs(y2 - y2prime) < shared_point_eps))
    return NO; /* shared point */

/* check for intersection (given by (0 < s < 1) && (0 < t < 1) */
  x2prime_x1prime = (double)(x2prime - x1prime);
  y2prime_y1prime = (double)(y2prime - y1prime);
  x2_x1           = (double)(x2 - x1);
  y2_y1           = (double)(y2 - y1);
  x1prime_x1      = (double)(x1prime - x1);
  y1prime_y1      = (double)(y1prime - y1);
  delta           = (x2prime_x1prime*y2_y1)     -(x2_x1*y2prime_y1prime);
  delta1          = (x2prime_x1prime*y1prime_y1)-(x1prime_x1*y2prime_y1prime);
  delta2          = ((-x1prime_x1) * y2_y1) + (x2_x1 * y1prime_y1);
  id_eps          = (fabs(x2prime_x1prime) + fabs(y2prime_y1prime)
                     + fabs(x2_x1) + fabs(y2_y1));
  if (ALMOST_PARALELL(delta)) {
#define POINTS_IDENTICAL(X1, Y1, X2, Y2)                               \
    ((fabs(X2-X1)/id_eps < 1.e-3) && (fabs(Y2-Y1)/id_eps < 1.e-3))
#define SIGN_OF(D) ((D) > 0? +1: ((D) < 0? -1: 0))
    if        (POINTS_IDENTICAL(x1, y1, x1prime, y1prime) ||
	       POINTS_IDENTICAL(x2, y2, x2prime, y2prime)) {
      if (SIGN_OF(x2_x1) == SIGN_OF(x2prime_x1prime) &&
	  SIGN_OF(y2_y1) == SIGN_OF(y2prime_y1prime))
	return YES;
    } else if (POINTS_IDENTICAL(x2, y2, x1prime, y1prime) ||
	       POINTS_IDENTICAL(x1, y1, x2prime, y2prime)) {
      if (SIGN_OF(-x2_x1) == SIGN_OF(x2prime_x1prime) &&
	  SIGN_OF(-y2_y1) == SIGN_OF(y2prime_y1prime))
	return YES;
    }
    return NO;
  }
  drev = 1. / delta;
  t = delta1 * drev;
  s = delta2 * drev;
  if ((t > 0 && t < 1) && (s > 0 && s < 1))
    return YES;
  else
    return NO;
}


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#define PI_HALF 1.5707963


/* RAW: needs to be optimized -- if we are looking for the ideal edge */
/* at a vertex, then we can use the stored values. */

float
ideal_edge_length(float fx, float fy, float lx, float ly,
                  float mdim, float mf, float ang)
{
#ifdef ANISOTROPIC_MESH
  float phi, tau, cp, sp;
  float xbar, ybar;

  if (options.anisotropic_mesh) {
    if (mdim == 0) { /* must compute */
      mdim  = get_mesh_cdim(fx, fy);
      mf    = get_mesh_2nd_dir_magn_factor(fx, fy);
      ang   = get_mesh_1st_dir_angle(fx, fy);
    }
    
    phi   = atan2(ly-fy, lx-fx) - ang;
    cp    = cos(phi);
    sp    = sin(phi);
    if (fabs(cp) < 0.0001)
      if (sp > 0) tau =  M_PI/2;
      else        tau = -M_PI/2;
    else
      tau   = atan2(sp, cp*mf);
    xbar  = mdim * cos(tau);
    ybar  = mdim * mf * sin(tau);
    
    return sqrt(xbar*xbar + ybar*ybar);
  } else {
#endif
    if (mdim != 0) { /* passed in -- had it from vertex */
      return mdim;
    } else {
      return get_mesh_cdim(fx, fy);
    }
#ifdef ANISOTROPIC_MESH
  }
#endif
}


/* ======================================================================== */
#if 0 /* DO NOT NEED IDEAL RADIUS ANY MORE */
float 
ideal_radius(VERTEX_REC *v1_p, VERTEX_REC *v2_p,
             float length, float ipx, float ipy)
{
  float temp, mdim1, mdim2;
  float idealr;

#ifdef ANISOTROPIC_MESH
  mdim1 = get_stored_cdim(v1_p);
  mdim2 = get_stored_cdim(v2_p);
  if (options.anisotropic_mesh) {
    float fx, lx, fy, ly;

    get_vertex_coords(v1_p, &fx, &fy);
    get_vertex_coords(v2_p, &lx, &ly);
    return AVRG(ideal_edge_length(fx, fy, ipx - fx, ipy - fy,
                                  mdim1, 
                                  get_stored_mf(v1_p),
                                  get_stored_angle(v1_p)),
                ideal_edge_length(lx, ly, ipx - lx, ipy - ly,
                                  mdim2, 
                                  get_stored_mf(v2_p),
                                  get_stored_angle(v2_p)));
  } else {
    idealr  = 0.5 * (mdim1 + mdim2);
    temp = options.upper_lim_on_ideal_radius * length;
    if (idealr > temp) 
      return temp;
    temp = LOWER_LIMIT_ON_IDEAL_RADIUS * length;
    if (idealr < temp)
      return temp; 
  }
  return idealr;
#else  /* !ANISOTROPIC_MESH */
  mdim1 = get_stored_cdim(v1_p);
  mdim2 = get_stored_cdim(v2_p);
  mdim = 0.5 * (mdim1 + mdim2);
#ifdef IDEAL_RADIUS_REDUCTION
  mdim *= IDEAL_RADIUS_REDUCTION;
#endif
  temp = options.upper_lim_on_ideal_radius * length;
  if (mdim > temp) 
    return temp;
  temp = LOWER_LIMIT_ON_IDEAL_RADIUS * length;
  if (mdim < temp)
    return temp;
  return (idealr = mdim);
#endif
}
#endif
/* ======================================================================== */

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void
ideal_point(VERTEX_REC *v1_p, VERTEX_REC *v2_p, float *ipx, float *ipy)
{
  float rf, rl, phi, xbar, ybar, rat;
  float ipxf, ipyf, ipxl, ipyl;
  float ecx, ecy, length;
  float fx, fy, lx, ly, gamma;
  
#ifdef ANISOTROPIC_MESH

#define KAPPA_FUNC_VAL(aKAPPA) (1/aKAPPA)
  
#define ANGLES()                                                             \
  {                                                                          \
    beta     = gamma - alpha;                                                \
    delta    = beta  + M_PI_2;                                               \
    if (delta < 0) delta += 2*M_PI;                                          \
    if (delta > 2*M_PI) delta -= 2*M_PI;                                     \
    g        = gamma - alpha;                                                \
    if (g < 0) g += 2*M_PI;                                                  \
    if        (g > (2*M_PI - alpha_lim) || g <= alpha_lim) {         /* A */ \
      phi =   M_PI_2 - (  M_PI_2 - delta) * kappa_f;                         \
    } else if (g > alpha_lim && g <= (M_PI - alpha_lim)) {           /* B */ \
      phi =   M_PI   - (  M_PI   - delta) * kappa_f;                         \
    } else if (g > (M_PI - alpha_lim) && g <= (M_PI + alpha_lim)){   /* C */ \
      phi = 3*M_PI_2 - (3*M_PI_2 - delta) * kappa_f;                         \
    } else if (g > (M_PI + alpha_lim) && g <= (2*M_PI - alpha_lim)){ /* D */ \
      if (delta < M_PI) phi = (delta) * kappa_f;                             \
      else              phi = 2*M_PI   - (2*M_PI  - delta) * kappa_f;        \
    } else {                                                                 \
      manip_error("bad sector");                                             \
    }                                                                        \
  }


  float beta, alpha, delta, cp, sp, tau, g;
  float mdim, mf, alpha_lim;
  float kappa_f;

  get_vertex_coords(v1_p, &fx, &fy);
  get_vertex_coords(v2_p, &lx, &ly);
  length = line_length(fx, fy, lx, ly);
  ecx    = 0.5 * (fx + lx);
  ecy    = 0.5 * (fy + ly);
  gamma  = atan2(ly - fy, lx - fx);
  if (gamma < 0) gamma +=  2*M_PI;

  if (options.anisotropic_mesh) {
    alpha_lim = AVRG(PE(mf_at_first), PE(mf_at_last));
    alpha_lim = atan2(2*alpha_lim, 1);
    
    /* first vertex */
    alpha    = PE(ang_at_first);
    if (alpha  < 0) alpha += 2*M_PI;
    mdim     = PE(cdim_at_first);
    mf       = PE(mf_at_first);
    kappa_f  = KAPPA_FUNC_VAL(mf);
    ANGLES();
    cp = cos(phi);
    sp = sin(phi);
    if (fabs(cp) < 0.0001)
      if (sp > 0) tau =  M_PI_2;
      else        tau = -M_PI_2;
    else
      tau   = atan2(sp, cp*mf);
    xbar = mdim * cos(tau);
    ybar = mdim * mf * sin(tau);
    rf   = sqrt(xbar * xbar  + ybar * ybar);
    rat = 1;
    if (rf > mf * length)
      rat = mf * length / rf;
    if (rf < mdim)
      rat = mdim / rf;  
    ipxf = fx + rat * (xbar * cos(alpha) - ybar * sin(alpha));
    ipyf = fy + rat * (xbar * sin(alpha) + ybar * cos(alpha));

    
    /* second vertex */
    alpha    = PE(ang_at_last);
    if (alpha  < 0) alpha += 2*M_PI;
    mdim     = PE(cdim_at_last);
    mf       = PE(mf_at_last);
    kappa_f  = KAPPA_FUNC_VAL(mf);
    ANGLES();
    cp = cos(phi);
    sp = sin(phi);
    if (fabs(cp) < 0.0001)
      if (sp > 0) tau =  M_PI_2;
      else        tau = -M_PI_2;
    else
      tau   = atan2(sp, cp*mf);
    xbar = mdim * cos(tau);
    ybar = mdim * mf * sin(tau);
    rl   = sqrt(xbar * xbar  + ybar * ybar);
    rat = 1;
    if (rl > mf * length)
      rat = mf * length / rl;
    if (rl < mdim)
      rat = mdim / rl; 
    ipxl = lx + rat * (xbar * cos(alpha) - ybar * sin(alpha));
    ipyl = ly + rat * (xbar * sin(alpha) + ybar * cos(alpha));
    rl   *=  rat;

    /* Resulting location */
    *ipx   = (ipxf * rl + ipxl * rf) / (rf + rl); 
    *ipy   = (ipyf * rl + ipyl * rf) / (rf + rl);
    /* Check */
    if (((*ipx - ecx)*(fy - ly) +  (*ipy - ecy)*(lx - fx)) < 0) {
      ideal_point(v1_p, v2_p, ipx, ipy);
      manip_error("bad ideal point");
    }
  } else {
#endif
    float cosphi, sinphi, rfrl_inv, shift;

    rf = PE(cdim_at_first);
    rl = PE(cdim_at_last);
   
#ifndef ANISOTROPIC_MESH
    get_vertex_coords(v1_p, &fx, &fy);
    get_vertex_coords(v2_p, &lx, &ly);
    length = line_length(fx, fy, lx, ly);
    ecx    = 0.5 * (fx + lx);
    ecy    = 0.5 * (fy + ly);
    gamma  = atan2(ly - fy, lx - fx);
#endif

    phi    = gamma + M_PI/2;
    cosphi = cos(phi);
    sinphi = sin(phi);

    /* first vertex */
    xbar = rf * cosphi;
    ybar = rf * sinphi;
    rat = 1;
    if (rf > options.upper_lim_on_ideal_radius * length)
      rat = options.upper_lim_on_ideal_radius * length / rf;
    if (rf < LOWER_LIMIT_ON_IDEAL_RADIUS * length)
      rat = LOWER_LIMIT_ON_IDEAL_RADIUS * length / rf; 
    ipxf  = ecx + rat * xbar;
    ipyf  = ecy + rat * ybar;
    rf   *=  rat;
    
    /* second vertex */
    xbar = rl * cosphi;
    ybar = rl * sinphi;
    rat = 1;
    if (rl > options.upper_lim_on_ideal_radius * length)
      rat = options.upper_lim_on_ideal_radius * length / rl;
    if (rl < LOWER_LIMIT_ON_IDEAL_RADIUS * length)
      rat = LOWER_LIMIT_ON_IDEAL_RADIUS * length / rl; 
    ipxl  = ecx + rat * xbar;
    ipyl  = ecy + rat * ybar;
    rl   *=  rat;

    /* Resulting location */
    rfrl_inv   = 1 / (rf + rl);
    shift      = max(RIGHT_SHIFT, rf * rfrl_inv);
    shift      = min(LEFT_SHIFT, shift);
    *ipx   = (ipxf * rl + ipxl * rf) * rfrl_inv;
    *ipy   = (ipyf * rl + ipyl * rf) * rfrl_inv;
    *ipx   += (shift - 0.5) * (lx - fx);
    *ipy   += (shift - 0.5) * (ly - fy);
#ifdef ANISOTROPIC_MESH
  }
#endif
}



 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
update_front(VERTEX_REC *selvertexp, float selvx, float selvy,
             EDGE_REC_P *edge1_p, EDGE_REC_P *edge2_p)
{
  EDGE_REC_P er[MAX_CONNECTIVITY];
  EDGE_REC_P ep;
  VERTEX_REC *v1_p = NULL, *v2_p = NULL, *v_p = NULL;
  BOOLEAN toremove;
  int n, k;
  sense sense;
  
  get_vert_references(selvertexp, &n, er);
  for (k = 0; k < n; k++) {
    ep = *(er+k);
    get_edge_vertex_p(ep, &v1_p, &v2_p);
    if (v1_p == selvertexp) 
      v_p = v2_p;
    else
      v_p = v1_p;
    toremove = NO;
    if (v_p == PE(v1_p)) {
      *edge1_p = ep;
      if (v1_p == PE(v1_p)) 
	sense = REVERSED_SENSE;
      else
        sense = AS_DEFINED_SENSE;
      toremove   = YES;
    } else if (v_p == PE(v2_p)) {
      *edge2_p = ep;
      if (v2_p == PE(v2_p)) 
	sense = REVERSED_SENSE;
      else
	sense = AS_DEFINED_SENSE;
      toremove   = YES;
    }
    if (toremove) {
      remove_edge_from_front(ep, sense);
    }
  }
  if (*edge1_p == NULL) {
    *edge1_p = add_edge(PE(v1_p), selvertexp, NULL);
    add_to_front(*edge1_p, AS_DEFINED_SENSE, 
		 line_length(selvx, selvy, FX, FY));
  }
  if (*edge2_p == NULL) {
    *edge2_p = add_edge(selvertexp, PE(v2_p), NULL);
    add_to_front(*edge2_p, AS_DEFINED_SENSE, 
		 line_length(selvx, selvy, LX, LY));
  }
}


static int
num_of_flanking_edges(VERTEX_REC *v,
                      VERTEX_REC *v1_to_match, VERTEX_REC *v2_to_match)
{
  int n, k;
  EDGE_REC_P er[MAX_CONNECTIVITY];
  BOOLEAN had_one = NO;
  VERTEX_REC *v1p = NULL, *v2p = NULL;

  get_vert_references(v, &n, er);
  for (k = 0; k < n; k++) {
    get_edge_vertex_p(*(er+k), &v1p, &v2p);
    if (v1p == v1_to_match || v2p == v1_to_match)
      if (had_one) return 2;
      else         had_one = YES;
    if (v1p == v2_to_match || v2p == v2_to_match)
      if (had_one) return 2;
      else         had_one = YES;
  }
  if (had_one)
    return 1;
  else
    return 0;
}
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

static VERTEX_REC *
can_use_intersected_edge(EDGE_REC_P intersected)
{
  VERTEX_REC *v1p = NULL, *v2p = NULL, *ret_vertex = NULL;
  float x1, y1, x2, y2, x, y;
  EDGE_REC_P isectswith = NULL;
  extern PULLED_EDGE_REC pulled_edge;
  BOOLEAN can_use_1 = NO, can_use_2 = NO;

  
  if (intersected != NULL) {
    get_edge_vertex_p(intersected, &v1p, &v2p);

    if (v1p != PE(v1_p) && v1p != PE(v2_p)) {
      get_vertex_coords(v1p, &x1, &y1);
      isectswith = NULL;
      if (no_intersections(v1p, x1, y1, NULL, &isectswith)) {can_use_1 = YES;}
    }

    if (v2p != PE(v1_p) && v2p != PE(v2_p)) {
      get_vertex_coords(v2p, &x2, &y2);
      isectswith = NULL;
      if (no_intersections(v2p, x2, y2, NULL, &isectswith)) {can_use_2 = YES;}
    }

    if (!can_use_1 && can_use_2) ret_vertex = v2p;
    if (!can_use_2 && can_use_1) ret_vertex = v1p;
    if (can_use_1 && can_use_2) {
      if (line_length(x1, y1, IPX, IPY) < line_length(x2, y2, IPX, IPY))
        ret_vertex = v1p;
      else
        ret_vertex = v2p;
    }
  }

  if (ret_vertex != NULL) {
    get_vertex_coords(ret_vertex, &x, &y);
    if ((FY - LY) * (x - 0.5*(FX + LX)) + (LX - FX) * (y - 0.5*(FY + LY)) <= 0)
      ret_vertex = NULL;        /* not in the "inside" half-plane */
  }
  
  return ret_vertex;
}



static void 
small_angle_check(VERTEX_REC **selvertexp,
                  EDGE_REC_P *edge1_p, EDGE_REC_P *edge2_p,
                  float *selvx, float *selvy)
{
  EDGE_REC_P isectswith_p;
  EDGE_REC_P er[MAX_CONNECTIVITY];
  float edge0, edge1, evec0, evec1, el, vecl;
  int n, k;
  VERTEX_REC *v1p = NULL, *v2p = NULL, *vp = NULL;
  float x, y, nx, ny, small_ang_cos, kappa;

  nx = -(LY-FY);
  ny =  (LX-FX);
#ifdef ANISOTROPIC_MESH
  if (options.anisotropic_mesh) {
    kappa = AVRG(PE(mf_at_first), PE(mf_at_last));
    small_ang_cos = (kappa / sqrt(kappa * kappa + 0.25));
  } else {
#endif
    small_ang_cos = SMALLANGCOS;
#ifdef ANISOTROPIC_MESH
  }
#endif
  *edge1_p     = NULL;
  *edge2_p     = NULL;
  isectswith_p = NULL;
  get_vert_references(PE(v1_p), &n, er);
  edge0 = *selvx - FX;
  edge1 = *selvy - FY;
  el    = line_length(*selvx, *selvy, FX,  FY);
  for (k = 0; k < n; k++) {
    get_edge_vertex_p(*(er+k), &v1p, &v2p);
    vp = NULL;
    if (v1p == PE(v1_p)) { 
      if (v2p != PE(v2_p))
	vp = v2p;
    } else {
      if (v1p == PE(v2_p)) {
	if (v2p != PE(v1_p)) 
	  vp = v2p;
      } else {
	vp = v1p;
      }
    }
    if (vp != NULL) {
      get_vertex_coords(vp, &x, &y);
      evec0 = x - FX;
      evec1 = y - FY;
      vecl  = line_length(x, y, FX,  FY);
      if (nx*evec0 + ny*evec1 > 0)
        if (edge0*evec0 + edge1*evec1 > small_ang_cos * vecl * el) {
          if (no_intersections(vp, x, y, *(er+k), &isectswith_p)) {
            *selvertexp = vp;
            *selvx = x;
            *selvy = y;
            *edge1_p = *(er+k);
            update_front(*selvertexp, *selvx, *selvy, edge1_p, edge2_p);
            goto exit;
          }
        }
    }
  }
  get_vert_references(PE(v2_p), &n, er);
  edge0 = *selvx - LX;
  edge1 = *selvy - LY;
  el    = line_length(*selvx, *selvy, LX,  LY);
  for (k = 0; k < n; k++) {
    get_edge_vertex_p(*(er+k), &v1p, &v2p);
    vp = NULL;
    if (v1p == PE(v1_p)) { 
      if (v2p != PE(v2_p))
	vp = v2p;
    } else {
      if (v1p == PE(v2_p)) {
	if (v2p != PE(v1_p)) 
	  vp = v2p;
      } else {
	vp = v1p;
      }
    }
    if (vp != NULL) {
      get_vertex_coords(vp, &x, &y);
      evec0 = x - LX;
      evec1 = y - LY;
      vecl  = line_length(x, y, LX,  LY);
      if (nx*evec0 + ny*evec1 > 0)
        if (edge0*evec0 + edge1*evec1 > small_ang_cos * vecl * el) {
          if (no_intersections(vp, x, y, *(er+k), &isectswith_p)) {
            *selvertexp = vp;
            *selvx = x;
            *selvy = y;
            *edge2_p = *(er+k);
            update_front(*selvertexp, *selvx, *selvy, edge1_p, edge2_p);
            goto exit;
          }
        }
    }
  }
 exit:
  ;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#define GET_NEXT_EDGE_TO_CONSIDER(theEPP, theSENSEP) \
   { \
     EDGE_REC *anEP; \
     sense aSENSE; \
      \
     first_front_edge(&anEP, &aSENSE); \
     while (anEP != NULL) { \
       if ((anEP == *(theEPP)) && (aSENSE == *(theSENSEP))) { \
         break; \
       } \
       next_front_edge(&anEP, &aSENSE); \
     } \
     next_front_edge(theEPP, theSENSEP); \
   }


static BOOLEAN 
last_resort(VERTEX_REC **svp, float *svx, float *svy,
            EDGE_REC_P *edge1_p, EDGE_REC_P *edge2_p,
            EDGE_REC_P *isw)
{
  float ecx, ecy, vx, vy;
  EDGE_REC_P e1p;
  EDGE_REC_P e2p;
  EDGE_REC_P ep;
  EDGE_REC_P er[MAX_CONNECTIVITY];
  VERTEX_REC *v1st = NULL, *v2nd = NULL, *sv = NULL, *vps[2] = {NULL, NULL};
  int ne = 0, m, k;
  sense sns;
  
  ecx  = 0.5 * (FX + LX); ecy  = 0.5 * (FY + LY);
  vx   = -(LY - FY); vy   = LX - FX;

  first_front_edge(&ep, &sns);
  e1p = *edge1_p; e2p = *edge2_p;
  sv  = *svp;
  while (ep != NULL) {
    get_edge_vertex_p(ep, &sv, &v2nd);
    if (sv != PE(v1_p) && sv != PE(v2_p)) {
      get_vertex_coords(sv, svx, svy);
      if ((vx * (*svx - ecx) + vy * (*svy - ecy)) > 0.0) {
        *isw = NULL;
        if (no_intersections(sv, *svx, *svy, NULL, isw)) {
          update_front(*svp = sv, *svx, *svy, edge1_p, edge2_p);
          return SUCCESS;
        }
      }
    }
    GET_NEXT_EDGE_TO_CONSIDER(&ep, &sns);
  }

  if (ep == NULL) return FAILURE;
  
  vps[0] = PE(v1_p); vps[1] = PE(v2_p);
  for (m = 0; m < 2; m++) {
    e1p = *edge1_p; e2p = *edge2_p;
    get_vert_references(vps[m], &ne, er);
    for (k = 0; k < ne; k++) {
      ep = *(er+k);
      get_edge_vertex_p(ep, &v1st, &v2nd);
      if (v1st == vps[m]) sv = v2nd;
      else                sv = v1st;
      get_vertex_coords(sv, svx, svy);
      *isw = NULL;
      if (vx * (*svx - ecx) + vy * (*svy - ecy) > 0.0) {
	if (no_intersections(sv, *svx, *svy, NULL, isw)) {
	  update_front(*svp = sv, *svx, *svy, edge1_p, edge2_p);
	  return SUCCESS;
	}
      }
    }
  }
  return FAILURE;
}

