/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/

#include "targe2cf.h"
#include "ttypes.h"

#ifdef Q_TREE

#include "utils.h"
#include "tops.h"
#include "subregs.h"
#include "mctlp.h"
#include "qtree.h"

#ifdef X_XFACE

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <X11/Intrinsic.h>

#define HAVE_BOOLEAN

#include "xiface.h"
#include "Ebcbezs3d.h"
#include "Eactset.h"

#define COLOR_ENCODED_QUADS
#   ifdef COLOR_ENCODED_QUADS
void draw_a_leaf(QTREE_REC *q);
void draw_the_leaves(QTREE_REC *q);
float get_mesh_cdim(float x, float y);
#   endif
#endif

extern OPTS_REC options; /* import global options */

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 /* PRIVATE */
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

QTREE_REC *qt_root = NULL;

static float qt_root_llx, qt_root_lly, qt_root_urx, qt_root_ury;

static int depth = 0;
static int total_of_leaves = 0;
static long q_rec_total = 0;


static float previous_at_x = -FLT_MAX, previous_at_y = -FLT_MAX;
static float stored_mf = 1, stored_angle = 0;

#ifdef X_XFACE
#   ifdef NEED_QUAD_GREP
static EPixel qt_color;

static GraphicObj *draw_quad(QTREE_REC *q);
static void draw_ctlp(float x, float y, float c);
static void undraw_quad(QTREE_REC *q);
static void undraw_qtree(void);
#   endif
#endif

#define IS_LEAF(q)        ((q)->c_dim >= 0.0)

#define HAS_CTLP(q)       ((q)->spec.leaf.have_ctl_p)

#ifdef ANISOTROPIC_MESH
#   define SET_HAVE_CTLP(theQ, theX, theY, theCDIM, theMF, theANG)  \
  {                                                                 \
    (theQ)->spec.leaf.have_ctl_p = YES;                             \
    (theQ)->spec.leaf.ctlp_at_x  = theX;                            \
    (theQ)->spec.leaf.ctlp_at_y  = theY;                            \
    (theQ)->c_dim                = theCDIM;                         \
    (theQ)->mf                   = theMF;                           \
    (theQ)->ang                  = theANG;                          \
  } 
#else
#   define SET_HAVE_CTLP(theQ, theX, theY, theCDIM, theMF, theANG)  \
  {                                                                 \
    (theQ)->spec.leaf.have_ctl_p = YES;                             \
    (theQ)->spec.leaf.ctlp_at_x  = theX;                            \
    (theQ)->spec.leaf.ctlp_at_y  = theY;                            \
    (theQ)->c_dim                = theCDIM;                         \
  }  
#endif

#ifdef ANISOTROPIC_MESH
#   ifdef NEED_PARENT			  
#      define SAVE_LEAF(q, tmp_q) \
  tmp_q.center_x             = q->center_x; \
  tmp_q.center_x             = q->center_x; \
  tmp_q.center_y             = q->center_y; \
  tmp_q.dimension            = q->dimension; \
  tmp_q.c_dim                = q->c_dim; \
  tmp_q.mf                   = q->mf; \
  tmp_q.ang                  = q->ang; \
  tmp_q.parent_q             = q->parent_q; \
  tmp_q.spec.leaf.have_ctl_p = q->spec.leaf.have_ctl_p; \
  tmp_q.spec.leaf.ctlp_at_x  = q->spec.leaf.ctlp_at_x; \
  tmp_q.spec.leaf.ctlp_at_y  = q->spec.leaf.ctlp_at_y
#   else
#      define SAVE_LEAF(q, tmp_q) \
  tmp_q.center_x             = q->center_x; \
  tmp_q.center_x             = q->center_x; \
  tmp_q.center_y             = q->center_y; \
  tmp_q.dimension            = q->dimension; \
  tmp_q.c_dim                = q->c_dim; \
  tmp_q.mf                   = q->mf; \
  tmp_q.ang                  = q->ang; \
  tmp_q.spec.leaf.have_ctl_p = q->spec.leaf.have_ctl_p; \
  tmp_q.spec.leaf.ctlp_at_x  = q->spec.leaf.ctlp_at_x; \
  tmp_q.spec.leaf.ctlp_at_y  = q->spec.leaf.ctlp_at_y
#   endif
#else
#   ifdef NEED_PARENT			  
#      define SAVE_LEAF(q, tmp_q) \
  tmp_q.center_x             = q->center_x; \
  tmp_q.center_x             = q->center_x; \
  tmp_q.center_y             = q->center_y; \
  tmp_q.dimension            = q->dimension; \
  tmp_q.c_dim                = q->c_dim; \
  tmp_q.parent_q             = q->parent_q; \
  tmp_q.spec.leaf.have_ctl_p = q->spec.leaf.have_ctl_p; \
  tmp_q.spec.leaf.ctlp_at_x  = q->spec.leaf.ctlp_at_x; \
  tmp_q.spec.leaf.ctlp_at_y  = q->spec.leaf.ctlp_at_y
#   else
#      define SAVE_LEAF(q, tmp_q) \
  tmp_q.center_x             = q->center_x; \
  tmp_q.center_x             = q->center_x; \
  tmp_q.center_y             = q->center_y; \
  tmp_q.dimension            = q->dimension; \
  tmp_q.c_dim                = q->c_dim; \
  tmp_q.spec.leaf.have_ctl_p = q->spec.leaf.have_ctl_p; \
  tmp_q.spec.leaf.ctlp_at_x  = q->spec.leaf.ctlp_at_x; \
  tmp_q.spec.leaf.ctlp_at_y  = q->spec.leaf.ctlp_at_y
#   endif
#endif

#define TURN_TO_BRANCH(q)      ((q)->c_dim = -1.0)

#define LEAF_DIM_LARGER(q, d)  ((q)->dimension*0.99 > d)

#define Q_SAFETY_MARGIN 1.5

#define IDENTICAL_CTL_POINTS(q, x, y, c)            \
  ((fabs((x)-(q)->spec.leaf.ctlp_at_x)+             \
    fabs((y)-(q)->spec.leaf.ctlp_at_y))/(c) < 0.01)

#ifdef NEED_PARENT
#ifdef __STDC__
#   define I_AM_A(WHO, Q) ((Q) == (Q)->parent_q->spec.branch.WHO##_leaf)
#else
#   define I_AM_A(WHO, Q) ((Q) == (Q)->parent_q->spec.branch.WHO/**/_leaf)
#endif
#endif

#define XY_IN_ROOT(X, Y) (((X) <= qt_root_urx) && ((Y) <= qt_root_ury) && \
			  ((X) >= qt_root_llx) && ((Y) >= qt_root_lly))

#define LEAF_AT_POS(WHERE, Q) ((Q)->spec.branch.WHERE##_leaf)

#ifdef NEED_PARENT
#define ADJACENT(WHERE, Q) LEAF_AT_POS(WHERE, (Q)->parent_q)
#endif

#define TAKE_CDIM(Q) { v = ((Q)->c_dim); minv = min(minv, v); nsamples++; }

#define DESCEND(aX, aY, aD) { \
   descent_x = (aX); \
   descent_y = (aY); \
   descent_d = (aD); \
   descent_c = q->c_dim; \
   qt_start_sm_d_descend(qt_root); }

#define RETURN_CDIM(aQ) {                                      \
			   if ((aQ)->c_dim == 0)               \
			     qt_smooth_cdim(aQ);               \
			   return (aQ)->c_dim;                 \
			 }
  
#ifdef STATIC
#  undef STATIC
#  define STATIC static
#else
#  define STATIC static  
#endif
STATIC void       qt_add_ctlp(QTREE_REC *q, float x, float y,
                              float dim, float mf, float ang);
#ifdef NEED_PARENT
STATIC QTREE_REC *qt_make_node(QTREE_REC *parent,
			       float cx, float cy, float dim);
#define QT_MAKE_NODE(aPARENT, aCX, aCY, aDIM) \
        qt_make_node((aPARENT), (aCX), (aCY), (aDIM))
#else
STATIC QTREE_REC *qt_make_node(float cx, float cy, float dim);
#define QT_MAKE_NODE(aPARENT, aCX, aCY, aDIM) \
        qt_make_node((aCX), (aCY), (aDIM))
#endif
STATIC void       qt_smooth_leafsize(QTREE_REC *q);
STATIC void       qt_leaf_to_branch(QTREE_REC *q);
STATIC void       qt_sm_d_descend(QTREE_REC *q);
STATIC void       qt_start_sm_d_descend(QTREE_REC *q);
STATIC void       qt_smooth_cdim(QTREE_REC *q);
STATIC void       qt_complete_and_smooth_cdims(void);
STATIC void       qt_traverse(QTREE_REC *q);
static float      qt_peek_cdim_at(float x, float y);
DBG(STATIC int        total_of_the_leaves(QTREE_REC *q);)

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

STATIC void 
qt_leaf_to_branch(QTREE_REC *q)
{
  QTREE_REC tmp_q;
  float d, dh, dq, c, mf = 1, ang = 0, x, y, cx, cy;

  SAVE_LEAF(q, tmp_q);
  TURN_TO_BRANCH(q);
  d   = q->dimension; dh  = 0.5 * d; dq = 0.25 * d;
  c   = tmp_q.c_dim;
#ifdef ANISOTROPIC_MESH
  mf  = tmp_q.mf;
  ang = tmp_q.ang;
#endif
  cx  = q->center_x; cy = q->center_y;
  LEAF_AT_POS(ll, q) = QT_MAKE_NODE(q, cx-dq, cy-dq, dh);
  LEAF_AT_POS(ul, q) = QT_MAKE_NODE(q, cx-dq, cy+dq, dh);
  LEAF_AT_POS(lr, q) = QT_MAKE_NODE(q, cx+dq, cy-dq, dh);
  LEAF_AT_POS(ur, q) = QT_MAKE_NODE(q, cx+dq, cy+dq, dh);
  if (HAS_CTLP(&tmp_q)) {
    x = tmp_q.spec.leaf.ctlp_at_x;
    y = tmp_q.spec.leaf.ctlp_at_y;
    if (x <= cx) 
      if (y <= cy) { SET_HAVE_CTLP(LEAF_AT_POS(ll, q), x, y, c, mf, ang); }
      else         { SET_HAVE_CTLP(LEAF_AT_POS(ul, q), x, y, c, mf, ang); }
    else 
      if (y <= cy) { SET_HAVE_CTLP(LEAF_AT_POS(lr, q), x, y, c, mf, ang); }
      else         { SET_HAVE_CTLP(LEAF_AT_POS(ur, q), x, y, c, mf, ang); }
  }
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 
STATIC void qt_add_ctlp(QTREE_REC *q, float x, float y,
                        float cdim, float mf, float ang)
{
 restart:
  ;
  if (IS_LEAF(q)) {
    if (HAS_CTLP(q)) {
      if (!IDENTICAL_CTL_POINTS(q, x, y, cdim)) {
	qt_leaf_to_branch(q);
	goto restart;
      } 
    } else {
      SET_HAVE_CTLP(q, x, y, cdim, mf, ang);
    }
  } else {
    if (x <= q->center_x)  
      if (y <= q->center_y)
        qt_add_ctlp(LEAF_AT_POS(ll, q), x, y, cdim, mf, ang);
      else
        qt_add_ctlp(LEAF_AT_POS(ul, q), x, y, cdim, mf, ang);
    else 
      if (y <= q->center_y)
        qt_add_ctlp(LEAF_AT_POS(lr, q), x, y, cdim, mf, ang);
      else
        qt_add_ctlp(LEAF_AT_POS(ur, q), x, y, cdim, mf, ang);
  }
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

STATIC QTREE_REC *
#ifdef NEED_PARENT
qt_make_node(QTREE_REC *parent, float cx, float cy, float dim)
#else
qt_make_node(float cx, float cy, float dim)
#endif
{
  QTREE_REC *q;

  if ((q = (QTREE_REC *)make_node(sizeof(QTREE_REC))) == NULL)
    alloc_error("qt_make_node");
  q->center_x             = cx;
  q->center_y             = cy; 
  q->dimension            = dim;
  q->c_dim                = 0.0;
#ifdef ANISOTROPIC_MESH
  q->mf                   = 1;
  q->ang                  = 0;
#endif
#ifdef NEED_PARENT
  q->parent_q             = parent;
#endif
  q->spec.leaf.have_ctl_p = NO;
  q->spec.leaf.ctlp_at_x  = 0;
  q->spec.leaf.ctlp_at_y  = 0;

  q_rec_total++;
  return q;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

static float descent_x, descent_y, descent_d, descent_c;

#if 0 /* DON'T NEED IT */
STATIC void 
qt_start_sm_d_descend(QTREE_REC *q)
{
  qt_sm_d_descend(q);
}
#endif /* DON'T NEED IT */



STATIC void 
qt_sm_d_descend(QTREE_REC *q)
{
 restart:
  if (IS_LEAF(q)) {
    if (LEAF_DIM_LARGER(q, descent_d)) {
      qt_leaf_to_branch(q);
      goto restart;
    }
  } else {
    if (descent_x <= q->center_x) 
      if (descent_y <= q->center_y)
	qt_sm_d_descend(LEAF_AT_POS(ll, q));
      else
	qt_sm_d_descend(LEAF_AT_POS(ul, q));
    else 
      if (descent_y <= q->center_y)
	qt_sm_d_descend(LEAF_AT_POS(lr, q));
      else
	qt_sm_d_descend(LEAF_AT_POS(ur, q));
  }
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#define INVALID_ANG 100
#ifdef ANISOTROPIC_MESH
#   define INCR_MF_ANG  {                                                   \
                           mf     += stored_mf;                             \
                           cosang = cos(stored_angle);                      \
                           sinang = sin(stored_angle);                      \
                           if (ang != INVALID_ANG)  {                       \
                             if (fabs(stored_angle - ang) < M_PI_2) {       \
                               ang = (ang + atan2(sinang, cosang)) * 0.5;   \
                             }                                              \
                           } else {                                         \
                             ang = atan2(sinang, cosang);                   \
                           }                                                \
                           /* printf(">   %f; ", stored_angle); */          \
                         }
#else
#   define INCR_MF_ANG
#endif 

STATIC void 
qt_smooth_cdim(QTREE_REC *q)
{
  float s = 0, x, v, dh, cx, cy, mf = 0, ang = INVALID_ANG;
  float  cosang = 0, sinang = 0; 
  int nsamples = 0;
  float minv = FLT_MAX;

  dh = q->dimension*0.5;
  cx = q->center_x; cy = q->center_y;
 once_more:
  ang = INVALID_ANG;
  dh *= 1.45;
  x  = cx+dh;
  if (x <= qt_root_urx) {
    if (cy+dh <= qt_root_ury)
      if ((v = qt_peek_cdim_at(x, cy+dh)) > 0) {
	s += v; minv = min(minv, v); INCR_MF_ANG; nsamples++; }
    if ((v = qt_peek_cdim_at(x, cy)) > 0)      {
      s += v; minv = min(minv, v); INCR_MF_ANG; nsamples++; }
    if (cy-dh >= qt_root_lly)
      if ((v = qt_peek_cdim_at(x, cy-dh)) > 0) {
	s += v; minv = min(minv, v); INCR_MF_ANG; nsamples++; }
  }
  x = cx-dh;
  if (x >= qt_root_llx) {
    if (cy-dh >= qt_root_lly)
      if ((v = qt_peek_cdim_at(x, cy-dh)) > 0) {
	s += v; minv = min(minv, v); INCR_MF_ANG; nsamples++; }
    if ((v = qt_peek_cdim_at(x, cy)) > 0)      {
      s += v; minv = min(minv, v); INCR_MF_ANG; nsamples++; }
    if (cy+dh <= qt_root_ury)
      if ((v = qt_peek_cdim_at(x, cy+dh)) > 0) {
	s += v; minv = min(minv, v); INCR_MF_ANG; nsamples++; }
  }
  x = cx;
  if (cy+dh <= qt_root_ury) 
    if ((v = qt_peek_cdim_at(x, cy+dh)) > 0)   {
      s += v; minv = min(minv, v); INCR_MF_ANG; nsamples++; }
  if (cy-dh >= qt_root_lly) 
    if ((v = qt_peek_cdim_at(x, cy-dh)) > 0)   {
      s += v; minv = min(minv, v); INCR_MF_ANG; nsamples++; }
  
  if (nsamples > 0) {
    q->c_dim = s/nsamples;	/* average */
    q->c_dim =
      /*
      minv * (1 + (q->c_dim - minv)/(q->c_dim + minv)) 
      */
      (q->c_dim + minv) * 0.5	/* bias towards small mesh size */
      ;
#ifdef ANISOTROPIC_MESH
    q->mf = mf/nsamples;
    q->ang = ang;
    if (q->ang < 0) q->ang += 2*M_PI;
    /* printf("\naverage %f\n", q->ang); */
#endif
  }  else {
    goto once_more;
  }
}


#if 0 /* DON'T NEED IT */
STATIC void 
qt_smooth_leafsize(QTREE_REC *q)
{
#ifdef NEED_PARENT
  if (q->parent_q) {
    x = q->parent_q->center_x; y = q->parent_q->center_y;
    d = q->parent_q->dimension; dist = d/2 + q->dimension;
    if (x-dist >= qt_root_llx) {
      if (y-dist >= qt_root_lly)
	if (q != ADJACENT(ul, q))
	  DESCEND(x-dist, y-dist, d);
      if (y+dist <= qt_root_ury)
	if (q != ADJACENT(lr, q))
	  DESCEND(x-dist, y+dist, d);
    }
    if (x+dist <= qt_root_urx) {
      if (y-dist >= qt_root_lly)
	if (q != ADJACENT(ul, q))
	  DESCEND(x+dist, y-dist, d);
      if (y+dist <= qt_root_ury)
	if (q != ADJACENT(ll, q))
	  DESCEND(x+dist, y+dist, d);
    }
  }
#endif
}
#endif /*  DON'T NEED IT  */

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

STATIC void qt_traverse(QTREE_REC *q)
{
  if (IS_LEAF(q)) {
    if (!HAS_CTLP(q)) 
      qt_smooth_cdim(q);
  } else {
    qt_traverse(LEAF_AT_POS(ll, q));
    qt_traverse(LEAF_AT_POS(ul, q));
    qt_traverse(LEAF_AT_POS(ur, q));
    qt_traverse(LEAF_AT_POS(lr, q));
  }
}



#if 0 /* DON'T NEED IT */
STATIC void 
qt_complete_and_smooth_cdims()
{
  qt_traverse(qt_root);
}
#endif /* DON'T NEED IT */

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 /* PUBLIC */
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
qt_make_root(float minx, float miny, float maxx, float maxy)
{
  float dim;
#ifdef X_XFACE
#   ifdef NEED_QUAD_GREP
  BOOLEAN success;
#   endif
#endif

  dim = max((maxx-minx), (maxy-miny));

#ifdef X_XFACE
#   ifdef NEED_QUAD_GREP
  if (options.X_xface)
    qt_color = ColorGetPixelFromString("white", &success);
#   endif
#endif
  qt_root = QT_MAKE_NODE(NULL, minx+(0.5*dim), miny+(0.5*dim), dim);
  qt_root_llx = minx;
  qt_root_lly = miny;
  qt_root_urx = qt_root_llx+dim;
  qt_root_ury = qt_root_lly+dim;
}



/* NOTE: the function sets during the invocation the values of the */
/* second principal direction magnification factor and the angle between */
/* the first principal axis and the x-axis.  If the coordinates on input */
/* to qt_2nd_dir_magn_factor_at and qt_1st_dir_angle_at agree with those */
/* saved, the stored values are returned without descent. */

#ifdef ANISOTROPIC_MESH
#  define STORE_MF_ANG()          \
   {                              \
     previous_at_x =       (x);   \
     previous_at_y =       (y);   \
     stored_mf     =  (cq->mf);   \
     stored_angle  = (cq->ang);   \
   }
                                     
#else
#  define STORE_MF_ANG()          \
   {                              \
     previous_at_x =         x;   \
     previous_at_y =         y;   \
     stored_mf    =          1;   \
     stored_angle =          0;   \
   }
#endif


float 
qt_leaf_cdim_at(float x, float y)
{
  QTREE_REC *cq = qt_root;

  while (cq) {
    if (x <= cq->center_x) 
      if (y <= cq->center_y) 
	if (!IS_LEAF(cq)){ cq = LEAF_AT_POS(ll, cq); }
	else             { if (cq->c_dim == 0) qt_smooth_cdim(cq);
                           STORE_MF_ANG();
			   return cq->c_dim;
                         }
      else 
	if (!IS_LEAF(cq)){ cq = LEAF_AT_POS(ul, cq); }
	else             { if (cq->c_dim == 0) qt_smooth_cdim(cq);
                           STORE_MF_ANG();
			   return cq->c_dim;
                         }
    else 
      if (y <= cq->center_y) 
	if (!IS_LEAF(cq)){ cq = LEAF_AT_POS(lr, cq); }
	else             { if (cq->c_dim == 0) qt_smooth_cdim(cq);
                           STORE_MF_ANG();
			   return cq->c_dim;
                         }
      else 
	if (!IS_LEAF(cq)){ cq = LEAF_AT_POS(ur, cq); }
	else             { if (cq->c_dim == 0) qt_smooth_cdim(cq);
                           STORE_MF_ANG();
			   return cq->c_dim;
                         }
  }
  manip_error("No leaf in quadtree at this location!?");
  return 0.0;
}


float 
qt_2nd_dir_magn_factor_at(float x, float y)
{
  if (previous_at_x != x || previous_at_y != y) 
    qt_leaf_cdim_at(x, y);
  return stored_mf;
}
    

float 
qt_1st_dir_angle_at(float x, float y)
{
  if (previous_at_x != x || previous_at_y != y) 
    qt_leaf_cdim_at(x, y);
  return stored_angle;
}
    

static float 
qt_peek_cdim_at(float x, float y)
{
  QTREE_REC *cq = qt_root;

  while (cq) {
    if (x <= cq->center_x) 
      if (y <= cq->center_y) 
	if (!IS_LEAF(cq)){ cq = LEAF_AT_POS(ll, cq); }
	else             { STORE_MF_ANG(); return cq->c_dim; }
      else 
	if (!IS_LEAF(cq)){ cq = LEAF_AT_POS(ul, cq); }
	else             { STORE_MF_ANG(); return cq->c_dim; }
    else 
      if (y <= cq->center_y) 
	if (!IS_LEAF(cq)){ cq = LEAF_AT_POS(lr, cq); }
	else             { STORE_MF_ANG(); return cq->c_dim; }
      else 
	if (!IS_LEAF(cq)){ cq = LEAF_AT_POS(ur, cq); }
	else             { STORE_MF_ANG(); return cq->c_dim; }
  }
  manip_error("No leaf in quadtree at this location!?");
  return 0.0;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void qt_extend_qtree(float x, float y, float dim, float mf, float ang)
{
  if (qt_root == NULL)
    qt_build();

  depth = 0;
  qt_add_ctlp(qt_root, x, y, dim, mf, ang);
#ifdef X_XFACE 
#   ifdef NEED_QUAD_GREP
#if 0
  if (options.X_xface)
    draw_ctlp(x, y, dim);
#endif 
#   endif
#endif
}



void 
qt_build()
{
  float minx, miny, maxx, maxy;
  float d;

  if (qt_root != NULL)
    manip_error("QT_ROOT no longer NULL");
  
  subregions_bbox(&minx, &miny, &maxx, &maxy);
  d = min((maxx-minx), (maxy-miny));
  d *= 0.025;
  qt_make_root(minx-d, miny-d, maxx+d, maxy+d);
}



void 
qt_prepare_for_use(void)
{
  /*qt_complete_and_smooth_cdims();*/ /* uncomment to smooth all */
  DBG(printf("Total of QTREE_RECs = %ld\n", q_rec_total);)
    DBG(total_of_leaves = 0;
	printf("total of leaves %d\n", total_of_the_leaves(qt_root));
	)
}

#ifdef X_XFACE

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#   ifdef NEED_QUAD_GREP
static GraphicObj *
draw_quad(QTREE_REC *q)
{
  GraphicObj *g_p;
  WCRec point;
  double zval = 0;

  point.x = q->center_x;
  point.y = q->center_y;

  if (options.show_on_Z == 1)
    zval = q->c_dim;
  else if (options.show_on_Z == 2)
    zval = q->mf;
  else if (options.show_on_Z == 3)
    zval = q->ang;
  
  point.z = zval;
  
  g_p = CreateMarker3D(&point);
  EASValsSetColor(qt_color);
  EASValsSetMType(CIRCLE_MARKER);
  EASValsSetMSize(0);
  EGWithMaskChangeAttributes(MSIZE_MASK | MTYPE_MASK | COLOR_MASK, g_p);

  return g_p;
}



static void 
undraw_quad(QTREE_REC *q)
{
  if (!IS_LEAF(q)) {
    undraw_quad(LEAF_AT_POS(ll, q));
    undraw_quad(LEAF_AT_POS(lr, q));
    undraw_quad(LEAF_AT_POS(ul, q));
    undraw_quad(LEAF_AT_POS(ur, q));
  }
  EMDeleteGraphics(age_model, q->g_rep);
}



static void 
undraw_qtree(void)
{
  undraw_quad(qt_root);
}



#if 0
static void 
draw_ctlp(float x, float y, float c)
{
  GraphicObj *g_p;
  int range;
  WCRec point;

  point.x = x;
  point.y = y;
  g_p = CreateMarker3D(&point);
  range = ColorFringeValueToRange(EMGetAssocFringeTable(age_model), c);
  EASValsSetColor(qt_color);
  EASValsSetMSize(2);
  EGWithMaskChangeAttributes(MSIZE_MASK | COLOR_MASK, g_p);
  EMAddGraphicsToModel(age_model, g_p);
} 
#endif

#ifdef COLOR_ENCODED_QUADS
void 
draw_leaves(void)
{
  draw_the_leaves(qt_root);
}


void 
draw_the_leaves(QTREE_REC *q)
{
  if (IS_LEAF(q)) {
    draw_a_leaf(q);
  } else {
    draw_the_leaves(LEAF_AT_POS(ll, q));
    draw_the_leaves(LEAF_AT_POS(lr, q));
    draw_the_leaves(LEAF_AT_POS(ul, q));
    draw_the_leaves(LEAF_AT_POS(ur, q));
  }
} 



void 
draw_a_leaf(QTREE_REC *q)
{
  EMAddGraphicsToModel(age_model, draw_quad(q));
}
#endif

#   endif /* NEED_QUAD_GREP */
#endif /* X_XFACE */


DBG(
STATIC int total_of_the_leaves(QTREE_REC *q)
{
  if (IS_LEAF(q)) {
    return total_of_leaves++;
  } else {
    total_of_the_leaves(LEAF_AT_POS(ll, q));
    total_of_the_leaves(LEAF_AT_POS(lr, q));
    total_of_the_leaves(LEAF_AT_POS(ul, q));
    total_of_the_leaves(LEAF_AT_POS(ur, q));
  }
} 
)

#endif /* ifdef Q_TREE */
 
