/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/

#include "targe2cf.h"
#include "ttypes.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h> 
#ifdef MSDOS
#include <io.h>
/*#include <alloc.h>*/
#endif

#include "targe2.h"
#include "tops.h"
#include "log.h"

#include "loops.h"
#include "subregs.h"
#include "vertex.h"
#include "edges.h"
#include "curve.h"
#include "cells.h"
#include "mctlp.h"
#include "triangle.h"
#include "input.h"
#include "mesh.h"


#ifdef X_XFACE
#include "xiface.h"
#endif

#ifdef DOS_GXFACE

#include "views.h"
#include "scr.h"
#include "status.h"

#endif


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 /* PRIVATE */
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#ifdef UNIX
#   undef HAVE_GETOPT
#endif
#ifdef MSDOS
#   undef HAVE_GETOPT
#endif

#ifdef HAVE_GETOPT
   extern int optind, opterr;
   extern char *optarg;
#  define get_cl_options getopt
#else
   int targe2_optind = 0, targe2_opterr = 1;
   char *targe2_optarg = NULL;
#define optarg targe2_optarg
#define optind targe2_optind
#define opterr targe2_opterr
#  define get_cl_options GetOpt
   int GetOpt(int argc, char * const argv[], const char *flags);
#endif

OPTS_REC options; /* this structure holds all options */

#ifdef MCTL_BY_FUNC
int mctl_func_start(void);
#endif

#define TO_LOG_FILE 1
#define TO_STDOUT   0
#define MAX_BANNER_LINE 79
#define NAME_BUF_LENGTH MAXFILENAMECHAR

static void check_in_out_file_names(int argc, char **argv);

#ifdef MSDOS
#   define FILE_ACCESSIBLE(f) (access((f), 0) == 0)
#else
#   define FILE_ACCESSIBLE(f) (file_exists(f))
#endif

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void init_targe2(int argc, char **argv)
{
  get_options(argc, argv);
  init_data_structs();
#ifdef X_XFACE 
  if (options.X_xface) {
    build_xiface(argc, argv);
    xiface_loop(YES, "TARGE2 (C) P.Krysl 1994");
    ESISetExitCode(EXIT_RETURN_FAILURE); /* Exit from targe2 by Elixir */
                                         /* exit point means failure w.r.t. */
                                         /* the outcome of mesh generation */
                                         /* up to the output_mesh() function */
  }
#endif 
#ifdef DOS_GXFACE
  if (options.DOS_gxface) {
/* RAW: initialization of graphics, copyright etc. */
	initviewsys ();
	copyright ();
	initscreen ();
  }
#endif
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void print_usage_and_exit(int argc, char **argv)
{
  printf("\nUsage: %s -i in_file_name [ options ]\n", argv[0]);
#define PRINT_LINE(line) printf("\n%s", line)
#define PRINT_LINE_WITH_DEFAULT_FLOAT(line,n) \
                      printf("\n%s (default is %f)", line, (n))
#define PRINT_LINE_WITH_DEFAULT_INT(line,n) \
  printf("\n%s (default is %d)", line, (n))
  PRINT_LINE("   Options -->");
  PRINT_LINE("-n no output (default is produce output)");
  PRINT_LINE("-f output format AFRO=1, MOS=2 (default 1)");
  PRINT_LINE("-o out_file (default file in_file_name.out)");
  PRINT_LINE("-v verbose (produce log, default is don't)");
  PRINT_LINE("-s silent (don't produce any output to controlling terminal)");
  PRINT_LINE_WITH_DEFAULT_FLOAT(
     "-m merge tolerance", INITIAL_MERGE_TOLERANCE);
  PRINT_LINE_WITH_DEFAULT_FLOAT(
     "-C cell-size/pulled-edge length ratio", INITIAL_CELL_SIZE_TO_CDIM);
  PRINT_LINE_WITH_DEFAULT_FLOAT(
     "-S maximum stretch of ideal-point placement", 
     UPPER_LIMIT_ON_IDEAL_RADIUS);
  PRINT_LINE_WITH_DEFAULT_FLOAT(
     "-e mesh control shape function exponent", INITIAL_EXPONENT);
  PRINT_LINE("-L log_file (default file ``" LOG_OUT "'')");
#ifdef Q_TREE
  PRINT_LINE("-q use quadtree to control mesh size");
#endif
  PRINT_LINE_WITH_DEFAULT_INT("-O smoother type ", LAPLACIAN_SMOOTHING);
  PRINT_LINE("-Q # improve mesh quality by removing bad-connectivity vertices (yes=1, no=0) \n            (default is no; included automatically for option -A)");
#ifdef ANISOTROPIC_MESH
  PRINT_LINE("-A anisotropic mesh (oriented) (default is isotropic)");
  PRINT_LINE("-Z show as Z coordinate (1=h, 2=kappa, 3=alpha, 4=mesh-size cross) (default is 1)");
  PRINT_LINE("-p Pischweitz constant (default is 1.0)");
  PRINT_LINE("-D do not smooth mesh  (default is: run Laplacian smoothing)");
#endif
#ifdef MCTL_BY_FUNC
  PRINT_LINE("-F # use analytical function mesh control (default is don't)");
#endif
#ifdef QUASI_ESTIMATOR
  PRINT_LINE("-E quasi-estimate mesh size (default is don't)- estimate in in_file_name.est");
#endif
#ifdef X_XFACE
  PRINT_LINE("-X use X Window System graphical interface (default is don't)");
  PRINT_LINE("-$ stop for each triangle in the X loop (default is don't)");
#endif
#ifdef DOS_GXFACE
  PRINT_LINE("-g use DOS graphic interface (default is don't)");
  PRINT_LINE("-b graphic driver path");
#endif
#ifdef UNIX
  PRINT_LINE_WITH_DEFAULT_FLOAT("-B hash-function beta", INITIAL_BETA);
#endif
  PRINT_LINE_WITH_DEFAULT_FLOAT(
     "-M magnification factor in grid rebuild", INITIAL_MAGFACTOR);
  printf("\n"); 
#undef PRINT_LINE
  exit(EXIT_RETURN_FAILURE);
} 

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void get_options(int argc, char **argv) 
{
  int  ret;
  BOOLEAN quality_enhancements_disallowed = NO;
#define PRINT_LINE(line) printf("\n%s", line)

#ifdef X_XFACE
#define OPTSX "X$"
#else
#define OPTSX 
#endif
#ifdef DOS_GXFACE
#define OPTSDG "gb:"
#else
#define OPTSDG 
#endif
#ifdef Q_TREE  
#define OPTSQ "q"
#else
#define OPTSQ 
#endif
#ifdef ANISOTROPIC_MESH
#define OPTSA "AZ:Dp:"
#else
#define OPTSA 
#endif
#ifdef MCTL_BY_FUNC 
#define OPTSF "F:"
#else
#define OPTSF 
#endif
#ifdef STRICT_MESH_SIZE_CONTROL
#define OPTSBANG "!"
#else
#define OPTSBANG 
#endif
#ifdef QUASI_ESTIMATOR
#define OPTSE "E"
#else
#define OPTSE
#endif
#ifdef UNIX
#define OPTSB "B:"
#else
#define OPTSB 
#endif
#ifdef UNIX
#define OPTSM "M:"
#else
#define OPTSM 
#endif

/* Concatenate all-season options with specific options */
#define OPTIONS "i:nf:o:L:svC:S:e:m:O:Q:"                   \
       OPTSX OPTSQ OPTSF OPTSE OPTSDG OPTSB OPTSM OPTSA OPTSBANG

  if (argc == 1)
    print_usage_and_exit(argc, argv);
  options.use_logging               = NO;
  options.log_file_name             = NULL;
  options.no_output                 = NO;
  options.output_format             = 1;
  options.out_file_name             = NULL;
  options.mesh_q_sigma              = INITIAL_MESH_QUALITY_SIGMA;
  options.mesh_q_tau                = INITIAL_MESH_QUALITY_TAU;
  options.mesh_q_omega              = INITIAL_MESH_QUALITY_OMEGA;
  options.merge_tolerance           = INITIAL_MERGE_TOLERANCE;
  options.cell_s_2_c_dim            = INITIAL_CELL_SIZE_TO_CDIM;
  options.const_c_dim               = INITIAL_CONSTMCDIM;
  options.upper_lim_on_ideal_radius = UPPER_LIMIT_ON_IDEAL_RADIUS;
  options.mctl_func_exponent        = INITIAL_EXPONENT;
  options.silent_as_a_tomb          = NO;
  options.qtree                     = NO;
  options.anisotropic_mesh          = NO;
  options.smooth_final_mesh         = LAPLACIAN_SMOOTHING;
  options.show_on_Z                 = 0;
  options.Pischweitz                = 1;
  options.mctl_by_func              = NO;
#ifdef X_XFACE
  options.X_xface                   = NO;
  options.stop_for_each             = NO;
#endif
#ifdef DOS_GXFACE
  options.DOS_gxface                = NO;
  options.gdriver_path              = NULL;
#endif
  options.produce_quasi_estimate    = NO;
#ifdef UNIX
  options.hash_beta                 = INITIAL_BETA;
#endif
  options.magnific_factor           = INITIAL_MAGFACTOR;
  options.quality_enhancements      = NO;
#ifdef STRICT_MESH_SIZE_CONTROL
  options.strict_mesh_size_control  = NO;
#endif
  optind = 0; opterr = 1;
  while((ret = get_cl_options(argc, argv, OPTIONS)) != EOF) {
    switch(ret) {
    case '?':
      print_usage_and_exit(argc, argv);
      break;
    case 'n':
      options.no_output     = YES;
      break;
    case 'f':
      options.output_format = (int)strtod(optarg, (char **)NULL);
      break;  
    case 'o':
      DUPLICATE_STRING(options.out_file_name, optarg);
      break;
    case 'i':
      if (optarg == NULL) {
        fprintf(stderr, "\nNo input file name given\n");
        exit(EXIT_RETURN_FAILURE);
      }
      DUPLICATE_STRING(options.in_file_name, optarg);
      break;
    case 'L':
      options.use_logging   = YES;
      DUPLICATE_STRING(options.log_file_name, optarg);
      break;
    case 's':
      options.silent_as_a_tomb = YES;
      options.use_logging = NO;
      break;
    case 'v':
      options.use_logging = YES;
      break;
#ifdef STRICT_MESH_SIZE_CONTROL
    case '!':
      options.strict_mesh_size_control = YES;
      break;
#endif
    case 'q':
      options.qtree = YES;
      break;
    case 'O':
      options.smooth_final_mesh = (int)strtod(optarg, (char **)NULL);
      break;
    case 'Q':
      options.quality_enhancements
        = ((int)strtod(optarg, (char **)NULL) == 1);
      quality_enhancements_disallowed = !options.quality_enhancements;
      break;
#ifdef ANISOTROPIC_MESH
    case 'A':
      options.anisotropic_mesh = YES;
      break;
    case 'D':
      options.smooth_final_mesh = NO;
      break;  
    case 'Z':
      options.show_on_Z = (int)strtod(optarg, (char **)NULL);
      break;
    case 'p':
      options.Pischweitz = strtod(optarg, (char **)NULL);
      break;    
#endif
#ifdef QUASI_ESTIMATOR
    case 'E':
      options.produce_quasi_estimate = YES;
      break;
#endif
    case 'C': 
      options.cell_s_2_c_dim = strtod(optarg, (char **)NULL);
      break;
    case 'S': 
      options.upper_lim_on_ideal_radius = strtod(optarg, (char **)NULL);
      break;
    case 'e':
      options.mctl_func_exponent = strtod(optarg, (char **)NULL);
      break;
    case 'm':
      options.merge_tolerance = strtod(optarg, (char **)NULL);
      break;
#ifdef X_XFACE
    case 'X':
      options.X_xface = YES;
      break;
    case '$':
      options.stop_for_each = YES;
      break;    
#endif
#ifdef DOS_GXFACE
    case 'b':
      DUPLICATE_STRING(options.gdriver_path, optarg);
    case 'g':
      options.DOS_gxface = YES;
      break;  
#endif
#ifdef MCTL_BY_FUNC
    case 'F':
      options.mctl_by_func = (int)strtod(optarg, (char **)NULL);
      break;
#endif
#ifdef UNIX
    case 'B': 
      options.hash_beta = strtod(optarg, (char **)NULL);
      break;
#endif
    case 'M': 
      options.magnific_factor = strtod(optarg, (char **)NULL);
      break;
    default:
      {
        char buf[132];

        if (ret != 0) {
          sprintf(buf, "No such option available: %c", ret);
          manip_error(buf);
        }
      }
      break;
    }
  }
  if (options.anisotropic_mesh)
      if (!quality_enhancements_disallowed)
          options.quality_enhancements = YES;
  
  check_in_out_file_names(argc, argv);
  if (options.use_logging)
    report_startup();

  options.remap_ratio               = 1.001/options.cell_s_2_c_dim;

#undef PRINT_LINE
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

static void check_in_out_file_names(int argc, char **argv)
{
  char line[NAME_BUF_LENGTH];
  
  if (options.in_file_name == NULL)
    print_usage_and_exit(argc, argv);
  
  if (!FILE_ACCESSIBLE(options.in_file_name)) {
    strcpy(line, options.in_file_name);
    strcat(line, IN_SUFFIX);
    if (!FILE_ACCESSIBLE(line)) {
      fprintf(stderr, "\nInaccessible input file %s (or %s)\n",
	      options.in_file_name, line);
      exit(EXIT_RETURN_FAILURE);
    } else {
      free(options.in_file_name);
      DUPLICATE_STRING(options.in_file_name, line);
    }
  }
  
  if (options.out_file_name == NULL) {
    strcpy(line, options.in_file_name);
    if (has_suffix(line, IN_SUFFIX)) 
      line[strlen(line)-strlen(IN_SUFFIX)] = '\0';
    switch (options.output_format) {
    case 3: 
      strcat(line, WASP_SUFFIX);
      break;
    case 2: 
      strcat(line, MOS_SUFFIX);
      break;
    default:
      strcat(line, OUT_SUFFIX);
      break;
    }
    DUPLICATE_STRING(options.out_file_name, line);
  }
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void post_banner(int where_to)
{
  int p;
  char *banner[] = {
#ifdef MSDOS
    "                Triangulation of ARbitrary GEometries in 2D,                ",
    "  (C)1993,1994 Petr Krysl, Email:pk@power2.fsv.cvut.cz, Phone:(42 2)332 4472",
    " ",
#else
    "  ==========================================================================",
    "            \"T\"riangulation of \"AR\"bitrary \"GE\"ometries in \"2\"D     ",
    "              TARGE2 v.1, Copyright (C) 1993,1994 Petr Krysl                ",
    "     TARGE2 comes with ABSOLUTELY NO WARRANTY; for details see the file     ",
    "   `targe2.wrn'. This is free software, and you are welcome to redistribute ",
    "    it under certain conditions; see the file ``targe2.use'' for details.   ",
    "  ==========================================================================",
#endif
  };

  for (p = 0; p < sizeof(banner)/sizeof(banner[0]); p++)
    if (where_to == TO_LOG_FILE) 
      log_to_file(NULL, banner[p]);
    else 
      printf("%s", banner[p]);
}
  
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

time_t t;
struct tm *tm_ptr;
static char *(short_month_names[]) =  {
  "Jan", "Feb", "Mar", "Apr", "May", "Jun",
  "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};
  
void get_timing(char *return_string)
{
  t = time(0);
  tm_ptr = localtime(&t);
  sprintf(return_string, " %s %d, %d [%02d:%02d:%02d]",
	  short_month_names[tm_ptr->tm_mon],
	  tm_ptr->tm_mday, tm_ptr->tm_year + 1900,
	  tm_ptr->tm_hour, tm_ptr->tm_min, tm_ptr->tm_sec);
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void report_startup()
{
  char buffer[NAME_BUF_LENGTH];
  char line[NAME_BUF_LENGTH];

  post_banner(TO_LOG_FILE);
  sprintf(line,       "   Input file : %s", options.in_file_name);
  log_to_file(NULL, line);
  log_to_file(NULL,   "   Options:");
  if (!options.no_output) {
    sprintf(line,     "   Output file: %s", options.out_file_name);
    log_to_file(NULL, line);
  } else
    log_to_file(NULL, "   Output of triangulation results suppressed.");
  if (options.qtree)
    log_to_file(NULL, "   Using quadtree to control mesh.");
#ifdef MCTL_BY_FUNC
  if (options.mctl_by_func) {
    sprintf(line,     "   Using analytic function #%d to control mesh.",
	    options.mctl_by_func);
    log_to_file(NULL, line);
  }
#endif
  if (options.produce_quasi_estimate)
    log_to_file(NULL, "   Producing quasi-estimate of gradation.");
  strcpy(line, "\n            Started:  ");
  get_timing(buffer);
  strcat(line, buffer);
  log_to_file(NULL, line);
}
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void run_targe2()
{
  input_data();
  generate();
  output_mesh();
#ifdef X_XFACE
  ESISetExitCode(EXIT_RETURN_SUCCESS);
#endif
}
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void cleanup_targe2()
{
  RAW;
  close_log_file();
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void init_data_structs()
{
  make_subregs_list();
  make_curve_list();
  make_vertex_list();
  make_edge_list();
  make_loops_list();
  make_mctlp_list();
  make_triangle_list();
  dummy_cells(); 
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void alloc_error(char *in_proc)
{
#ifdef DOS_GXFACE
  if (options.DOS_gxface) {
/* RAW: fatal error */
    Clear_status(1);
    Print_status(1,ERRMSG1);
    Status_loop ();
    closegraph();
  } else {
    fprintf(stderr, "\nCould not allocate memory in %s.\n", in_proc);
    exit(EXIT_RETURN_FAILURE);
  }
#else
  fprintf(stderr, "\nCould not allocate memory in %s.\n", in_proc);
#   ifdef X_XFACE
  if (options.X_xface) {
    xiface_loop(YES, "MEMORY ALLOCATION ERROR.");
  }
#   else
  exit(EXIT_RETURN_FAILURE);
#   endif
#endif /* DOS_GXFACE */
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void generate()
{
#ifdef DOS_GXFACE
  char str[80];
#endif
  mesh_region();
#ifdef X_XFACE
  if (options.X_xface) {
    xiface_loop(YES, "FINISHED.");
  }
#endif
#ifdef DOS_GXFACE
  if (options.DOS_gxface) {
/* RAW: good place to ask for acceptance/rejection of the generated mesh */
	sprintf(str,END_FMT1,total_of_triangles(),total_of_vertices());
	Print_status(2,str);
	PrintMenuLine(MLINE2);
	Exit_loop ();
	closegraph();
  }
#endif  
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#ifndef HAVE_GETOPT /* if we have no getopt function */

int GetOpt(int argc, char * const argv[], const char *flags)
{
  static int started = NO;
  static const char *current_flag;
  int i, j, retval;

  if (!started) {
    current_flag  = flags;
    started       = YES;
  }
  if (*current_flag == '\0') {
    started = NO;
    return EOF;
  }
  
  for (i = 1; i < argc; i++) {
    if ((int)argv[i][0] == (int)'-') {
      j = 0;
      while (argv[i][j] != '\0') {
	if ((int)argv[i][j] == (int)(*current_flag)) {
	  retval = (int)(*current_flag);
	  if ((int)(*(current_flag+1)) == (int)':') {
	    optarg = argv[i+1];
	    current_flag++;
	  }
	  current_flag++;
	  return retval;
	}
	j++;
      }
    }
  }

  if (*(++current_flag) == '\0') {
    started = NO;
    return EOF;
  } else 
    return 0;
}

#endif

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

BOOLEAN has_suffix(char *f, char *s)
{
  int lf, ls, i;

  if (!f || !s)
    return NO;
  
  lf = strlen((f)); ls = min(strlen((s)), lf);
  for (i = 0; i < ls; i++)
    if (toupper((f)[lf-i]) != toupper((s)[ls-i]))
      return NO;
  return YES;
}


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#ifdef UNIX
BOOLEAN file_exists(char *f)
{
  struct stat statbuf;

  return (stat(f, &statbuf) == 0);
}
#endif

 
