/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/

#include "targe2cf.h"
#include "ttypes.h"
#include "triangle.h"
#include "edges.h"
#include "vertex.h"
#include "subregs.h"
#include "mctlp.h"
#include "tops.h"

#ifdef X_XFACE

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <X11/Intrinsic.h>

#define HAVE_BOOLEAN

#include "xiface.h"

void 
draw_tri(TRIANGLE_REC *r);

#endif

#ifdef DOS_GXFACE
#   define GRAPHICS_REC void
void 
draw_tri(TRIANGLE_REC *r);

#include "views.h"
#include "scr.h"
#include "status.h"

#endif

extern OPTS_REC options; /* import global options */

TRIANGLE_LIST triangle_list;

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
make_triangle_list()
{
  triangle_list = MAKE_LIST();
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

long int 
total_of_triangles()
{
  return LIST_LENGTH(triangle_list);
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

long int 
total_of_bad_triangles(void)
{
  unsigned long bad_ones = 0;
  float x1, y1, x2, y2, x3, y3;
  TRIANGLE_REC *t;
  VERTEX_REC_P v1;
  VERTEX_REC_P v2;
  VERTEX_REC_P v3;
  VERTEX_REC_P tmp_v1 = NULL;
  VERTEX_REC_P tmp_v2 = NULL;
  BOOLEAN good;
  int type;
  
  t = (TRIANGLE_REC *)GET_LIST_NEXT(triangle_list, NULL);
  while (t != NULL) {
    get_edge_vertex_p(t->edge_1, &tmp_v1, &tmp_v2);
    v1 = tmp_v1, v2 = tmp_v2;
    get_vertex_coords(v1, &x1, &y1);
    get_vertex_coords(v2, &x2, &y2);
    get_edge_vertex_p(t->edge_2, &tmp_v1, &tmp_v2);
    if (tmp_v1 == v1 || tmp_v1 == v2)
      v3 = tmp_v2;
    if (tmp_v2 == v1 || tmp_v2 == v2)
      v3 = tmp_v1;
    get_vertex_coords(v3, &x3, &y3);
    test_triangle_shape(x1, y1, x2, y2, x3, y3, &good, &type);
    if (!good)
      bad_ones++;
    t = (TRIANGLE_REC *)GET_LIST_NEXT(triangle_list, t);
  }
  return bad_ones;
}


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 /* add_triangle  */

BOOLEAN 
add_triangle(EDGE_REC *edge_1, EDGE_REC *edge_2, EDGE_REC *edge_3,
	     SUBREG_REC *subreg)
{
  TRIANGLE_REC *r;

  MAKE_AND_ADD_TO_TAIL_VAL(triangle_list, r, TRIANGLE_REC,
			    "add_triangle", 0);
  r->subreg    = subreg;
  r->edge_1    = edge_1;
  r->edge_2    = edge_2;
  r->edge_3    = edge_3;
  return YES;
}


void
destroy_triangle(TRIANGLE_REC *t)
{
  DELETE_LIST_NODE(triangle_list, t);
}


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
test_triangle_shape(float x1, float y1,
		    float x2, float y2,
		    float x3, float y3,
		    BOOLEAN *good, int *type)
{
  float sidea, sideb, sidec, temp1;
  float sigma, tau, omega;
  float sidesum, area, inscribedr, circumscribedr, minside, maxside;

  sidea = line_length(x1, y1, x2, y2);
  sideb = line_length(x3, y3, x2, y2);
  sidec = line_length(x1, y1, x3, y3);
  sidesum = 0.5 * (sidea + sideb + sidec);
  area    = sidesum*(sidesum - sidea)*(sidesum - sideb)*(sidesum - sidec);
  area    = sqrt(area);
  minside = min(sidea, sideb); minside = min(minside, sidec);
  maxside = max(sidea, sideb); maxside = max(maxside, sidec);
  temp1   = sidea * sideb * sidec;
  circumscribedr = temp1 * 0.25 / (area + 0.000001*temp1);
  inscribedr     = area / sidesum;
  sigma          = maxside / (inscribedr + 0.000001*maxside);
  omega          = circumscribedr / maxside;
  tau            = maxside / minside;

  if (sigma > options.mesh_q_sigma) {
    *good = NO;
    if (tau > options.mesh_q_tau)
      *type = ACUTE;
    else
      *type = OBTUSE;
  } else {
    *good = YES;
    *type = GOOD;
  }
}


#ifdef X_XFACE			/* <<<<<<<<<<<< X_XFACE */

EPixel bad_shape_color;

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

FPNum minv_in_table = 0, maxv_in_table = 1;

void 
draw_triangles(void)
{
  TRIANGLE_REC *t;
  BOOLEAN success;
  FPNum delta;

  bad_shape_color = ColorGetPixelFromString("red", &success);
  if (!success)
    bad_shape_color = 0x0;
  
  ColorFringesMinMax(EMGetAssocFringeTable(ESIModel ()),
		     &minv_in_table, &maxv_in_table);
  delta = ColorFringesDelta(EMGetAssocFringeTable(ESIModel ()));
  minv_in_table += delta; maxv_in_table -= delta;
  
  t = (TRIANGLE_REC *)GET_LIST_NEXT(triangle_list, NULL);
  while (t != NULL) {
    draw_tri(t);
    t = (TRIANGLE_REC *)GET_LIST_NEXT(triangle_list, t);
  }
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
draw_tri(TRIANGLE_REC *r)
{
  VERTEX_REC_P v1;
  VERTEX_REC_P v2;
  VERTEX_REC_P v3;
  VERTEX_REC_P tmp_v1 = NULL;
  VERTEX_REC_P tmp_v2 = NULL;
  float x1, y1, x2, y2, x3, y3, xc, yc;
#ifdef FILL_TRIANGLES
  float val1, val2, val3;
#endif
#ifndef DONT_DRAW_TRIANGLES
  BOOLEAN good;
  int type;
  GraphicObj *p;
  WCRec pnts[3] = {0, 0, 0, 0, 0, 0, 0, 0, 0};
#endif

  get_edge_vertex_p(r->edge_1, &tmp_v1, &tmp_v2);
  v1 = tmp_v1, v2 = tmp_v2;
  get_vertex_coords(v1, &x1, &y1);
  get_vertex_coords(v2, &x2, &y2);
  get_edge_vertex_p(r->edge_2, &tmp_v1, &tmp_v2);
  if (tmp_v1 == v1 || tmp_v1 == v2)
    v3 = tmp_v2;
  if (tmp_v2 == v1 || tmp_v2 == v2)
    v3 = tmp_v1;
  get_vertex_coords(v3, &x3, &y3);
#define SHRINK 0.85
  xc = 0.3333 * (x1+x2+x3);
  yc = 0.3333 * (y1+y2+y3);
  x1 = xc + SHRINK * (x1 - xc);
  x2 = xc + SHRINK * (x2 - xc);
  x3 = xc + SHRINK * (x3 - xc);
  y1 = yc + SHRINK * (y1 - yc);
  y2 = yc + SHRINK * (y2 - yc);
  y3 = yc + SHRINK * (y3 - yc);
#ifdef FILL_TRIANGLES
  val1 = get_stored_cdim(v1);
  val2 = get_stored_cdim(v2);
  val3 = get_stored_cdim(v3);
  xc   = 0.3333 * (val1+val2+val3);
  val1 = xc + SHRINK * (val1 - xc); val1 = 1./val1;
  val2 = xc + SHRINK * (val2 - xc); val2 = 1./val2;
  val3 = xc + SHRINK * (val3 - xc); val3 = 1./val3;
  pnts[0].x = x1;
  pnts[0].y = y1;
  pnts[1].x = x2;
  pnts[1].y = y2;
  pnts[2].x = x3;
  pnts[2].y = y3;
  p = CreateTriangle3D(pnts);
  EMAddGraphicsToModel(ESIModel (), p);
#else
#ifndef DONT_DRAW_TRIANGLES
  test_triangle_shape(x1, y1, x2, y2, x3, y3, &good, &type);
#ifdef LINE_TRI_GRAPHICS
  ModelAddGraphicsToModel(ESIModel (), p = CreateLine(x1, y1, x2, y2));
  if (!good)
    LineSetAttributes(p, COLOR_MASK | WIDTH_MASK, 0, 2, bad_shape_color);
  ModelAddGraphicsToModel(ESIModel (), p = CreateLine(x1, y1, x3, y3));
  if (!good)
    LineSetAttributes(p, COLOR_MASK | WIDTH_MASK, 0, 2, bad_shape_color);
  ModelAddGraphicsToModel(ESIModel (), p = CreateLine(x3, y3, x2, y2));
  if (!good)
    LineSetAttributes(p, COLOR_MASK | WIDTH_MASK, 0, 2, bad_shape_color);
#else
  {
    FPNum v;
    WCRec pnts[3] = {x1, y1, 0, x2, y2, 0, x3, y3, 0}; 
    
    if (!good)      v = maxv_in_table;
    else            v = minv_in_table;
    EMAddGraphicsToModel(ESIModel (), (p = CreateTriangleWD3D(pnts, v, v, v)));
  }
#endif
#endif
#endif
}

#endif  /* <<<<<<<<<<<< X_XFACE */

#ifdef DOS_GXFACE /* <<<<<<<<<<<< DOS_GXFACE */

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
draw_triangles()
{
  TRIANGLE_REC *t;
  
  t = (TRIANGLE_REC *)GET_LIST_NEXT(triangle_list, NULL);
  while (t != NULL) {
    draw_tri(t);
    t = (TRIANGLE_REC *)GET_LIST_NEXT(triangle_list, t);
  }
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#define GOOD   0
#define ACUTE  1
#define OBTUSE 2

void 
draw_tri(TRIANGLE_REC *r)
{
  VERTEX_REC_P v1;
  VERTEX_REC_P v2;
  VERTEX_REC_P v3;
  VERTEX_REC_P tmp_v1;
  VERTEX_REC_P tmp_v2;
  float x1, y1, x2, y2, x3, y3, xc, yc;
  float val1, val2, val3;
  BOOLEAN good;
  int type;
  GRAPHICS_REC *p;

  get_edge_vertex_p(r->edge_1, &tmp_v1, &tmp_v2);
  v1 = tmp_v1, v2 = tmp_v2;
  get_vertex_coords(v1, &x1, &y1);
  get_vertex_coords(v2, &x2, &y2);
  get_edge_vertex_p(r->edge_2, &tmp_v1, &tmp_v2);
  if (tmp_v1 == v1 || tmp_v1 == v2)
    v3 = tmp_v2;
  if (tmp_v2 == v1 || tmp_v2 == v2)
    v3 = tmp_v1;
  get_vertex_coords(v3, &x3, &y3);
#define SHRINK 0.85
  xc = 0.3333 * (x1+x2+x3);
  yc = 0.3333 * (y1+y2+y3);
  x1 = xc + SHRINK * (x1 - xc);
  x2 = xc + SHRINK * (x2 - xc);
  x3 = xc + SHRINK * (x3 - xc);
  y1 = yc + SHRINK * (y1 - yc);
  y2 = yc + SHRINK * (y2 - yc);
  y3 = yc + SHRINK * (y3 - yc);
#ifdef FILL_TRIANGLES
  val1 = get_stored_cdim(v1);
  val2 = get_stored_cdim(v2);
  val3 = get_stored_cdim(v3);
  xc   = 0.3333 * (val1+val2+val3);
  val1 = xc + SHRINK * (val1 - xc); val1 = 1./val1;
  val2 = xc + SHRINK * (val2 - xc); val2 = 1./val2;
  val3 = xc + SHRINK * (val3 - xc); val3 = 1./val3;
 /* RAW: draw the triangle as a filled color encoded polygon */
  setfillstyle(EMPTY_FILL,WHITE);
  Fillpolyview (x1,y1,x2,y2,x3,y3);
#else
#ifndef DONT_DRAW_TRIANGLES
  test_triangle_shape(x1, y1, x2, y2, x3, y3, &good, &type);
  if (!good)  {
    setfillstyle(SOLID_FILL,BAD_SHAPE_COLOR);
    Fillpolyview (x1,y1,x2,y2,x3,y3);
  }  else  {
    setfillstyle(SOLID_FILL,GOOD_SHAPE_COLOR);
    Fillpolyview (x1,y1,x2,y2,x3,y3);
  }
#endif
#endif
}

#endif  /* <<<<<<<<<<<< DOS_GXFACE */


#ifdef QUASI_ESTIMATOR
 
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
produce_mctlts(FILE *stream)
{
  int i;
  TRIANGLE_REC *t;
  VERTEX_REC_P v1;
  VERTEX_REC_P v2;
  VERTEX_REC_P v3;
  VERTEX_REC_P tmp_v1;
  VERTEX_REC_P tmp_v2;
  float x1, y1, x2, y2, x3, y3;
  float a[8];
  int numlines;

  numlines = total_of_triangles();
  fwrite(&numlines, sizeof(int), 1, stream);
  
  t = (TRIANGLE_REC *)GET_LIST_NEXT(triangle_list, NULL);
  i = 0;
  while (t != NULL) {
    get_edge_vertex_p(t->edge_1, &tmp_v1, &tmp_v2);
    v1 = tmp_v1, v2 = tmp_v2;
    get_vertex_coords(v1, &x1, &y1);
    get_vertex_coords(v2, &x2, &y2);
    get_edge_vertex_p(t->edge_2, &tmp_v1, &tmp_v2);
    if (tmp_v1 == v1 || tmp_v1 == v2)
      v3 = tmp_v2;
    if (tmp_v2 == v1 || tmp_v2 == v2)
      v3 = tmp_v1;
    get_vertex_coords(v3, &x3, &y3);
    a[0] = x1;
    a[1] = y1;
    a[2] = x2;
    a[3] = y2;
    a[4] = x3;
    a[5] = y3;
    a[6] = get_mesh_cdim(x1, y1);
    a[7] = get_mesh_cdim(x2, y2);
    a[8] = get_mesh_cdim(x3, y3);
    fwrite(a, sizeof(float), 9, stream);
    t = (TRIANGLE_REC *)GET_LIST_NEXT(triangle_list, t);
  }
}

#endif


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
dump_triangles_in_afro_style(FILE *fp)
{
  TRIANGLE_REC *t;
  int pid, mid, sid;
  long tid, v1id, v2id, v3id;
  float th, x1, y1, x2, y2, x3, y3;
  char *group;
  VERTEX_REC_P v1;
  VERTEX_REC_P v2;
  VERTEX_REC_P v3;
  VERTEX_REC_P e1v1 = NULL;
  VERTEX_REC_P e1v2 = NULL;
  VERTEX_REC_P e2v1 = NULL;
  VERTEX_REC_P e2v2 = NULL;
  VERTEX_REC_P e3v1 = NULL;
  VERTEX_REC_P e3v2 = NULL;
  

  fprintf(fp, "*TR3N\n");
  
  tid = 0;
  t = (TRIANGLE_REC *)GET_LIST_NEXT(triangle_list, NULL);
  while (t != NULL) {
    subreg_data(t->subreg, &pid, &mid, &sid, &th, &group);
    get_edge_vertex_p(t->edge_1, &e1v1, &e1v2);
    get_edge_vertex_p(t->edge_2, &e2v1, &e2v2);
    get_edge_vertex_p(t->edge_3, &e3v1, &e3v2);
    if      (e1v1 == e2v1 || e1v1 == e2v2) v1 = e1v1;
    else if (e1v2 == e2v1 || e1v2 == e2v2) v1 = e1v2;
    if      (e1v1 == e3v1 || e1v1 == e3v2) v3 = e1v1;
    else if (e1v2 == e3v1 || e1v2 == e3v2) v3 = e1v2;
    if      (e2v1 == e3v1 || e2v1 == e3v2) v2 = e2v1;
    else if (e2v2 == e3v1 || e2v2 == e3v2) v2 = e2v2;
    vertex_data(v1, &v1id, &x1, &y1);
    vertex_data(v2, &v2id, &x2, &y2);
    vertex_data(v3, &v3id, &x3, &y3);
    fprintf(fp, "%12lu  %12lu   %12lu   %12lu   %4d   %12.6f  %4d   %s\n",
	        ++tid, v1id, v2id, v3id, mid, th, sid, group);
    t = (TRIANGLE_REC *)GET_LIST_NEXT(triangle_list, t);
  }
}



void 
dump_triangles_in_mos_style(FILE *fp)
{
  TRIANGLE_REC *t;
  long tid;

  /* IMPORTANT NOTICE: */
  /* The routine uses the /cell/ field of the edge as the edge id. */
  /* This value must have been set by dump_edges_in_mos_style(); */
  /* */
  tid = 0;
  t = (TRIANGLE_REC *)GET_LIST_NEXT(triangle_list, NULL);
  while (t != NULL) {
    fprintf(fp, "%12lu     %12lu %12lu %12lu     %d\n",
            ++tid,
            t->edge_1->cell, t->edge_2->cell, t->edge_3->cell,
            t->subreg->subreg_id);
    t = (TRIANGLE_REC *)GET_LIST_NEXT(triangle_list, t);
  }
}

void 
dump_triangles_in_wasp_style(FILE *fp)
{
  TRIANGLE_REC *t;
  int pid, mid, sid;
  long tid, v1id, v2id, v3id;
  float th, x1, y1, x2, y2, x3, y3;
  char *group;
  VERTEX_REC_P v1;
  VERTEX_REC_P v2;
  VERTEX_REC_P v3;
  VERTEX_REC_P e1v1 = NULL;
  VERTEX_REC_P e1v2 = NULL;
  VERTEX_REC_P e2v1 = NULL;
  VERTEX_REC_P e2v2 = NULL;
  VERTEX_REC_P e3v1 = NULL;
  VERTEX_REC_P e3v2 = NULL;
  

  fprintf (fp, "ELEM\n");
  fprintf (fp, "type\n");
  tid = 0;
  t = (TRIANGLE_REC *)GET_LIST_NEXT(triangle_list, NULL);
  while (t != NULL) {
    subreg_data(t->subreg, &pid, &mid, &sid, &th, &group);
    get_edge_vertex_p(t->edge_1, &e1v1, &e1v2);
    get_edge_vertex_p(t->edge_2, &e2v1, &e2v2);
    get_edge_vertex_p(t->edge_3, &e3v1, &e3v2);
    if      (e1v1 == e2v1 || e1v1 == e2v2) v1 = e1v1;
    else if (e1v2 == e2v1 || e1v2 == e2v2) v1 = e1v2;
    if      (e1v1 == e3v1 || e1v1 == e3v2) v3 = e1v1;
    else if (e1v2 == e3v1 || e1v2 == e3v2) v3 = e1v2;
    if      (e2v1 == e3v1 || e2v1 == e3v2) v2 = e2v1;
    else if (e2v2 == e3v1 || e2v2 == e3v2) v2 = e2v2;
    vertex_data(v1, &v1id, &x1, &y1);
    vertex_data(v2, &v2id, &x2, &y2);
    vertex_data(v3, &v3id, &x3, &y3);
    fprintf(fp, "%12lu  %12lu   %12lu   %12lu   %4d   %12.6f  %4d   %s\n",
	        ++tid, v1id, v2id, v3id, mid, th, sid, group);
    t = (TRIANGLE_REC *)GET_LIST_NEXT(triangle_list, t);
  }
  fprintf (fp, "&\n");
}



SUBREG_REC *
delete_triangles_and_edges_around_vertex(VERTEX_REC *v_p);

static int
num_of_matches(TRIANGLE_REC *t, int n, EDGE_REC *er[MAX_CONNECTIVITY]);


SUBREG_REC *
delete_triangles_and_edges_around_vertex(VERTEX_REC *v_p)
{
  SUBREG_REC *sr = NULL;
  TRIANGLE_REC *t = NULL;
  EDGE_REC *er[MAX_CONNECTIVITY];
  int i, m, n;

  get_vert_references(v_p, &n, er);

 restart:
  t = (TRIANGLE_REC *)GET_LIST_NEXT(triangle_list, t);
  while (t != NULL) {
    m = num_of_matches(t, n, er);
    if (m == 2) {
      sr = t->subreg;
      destroy_triangle(t);
      goto restart;
    }
    t = (TRIANGLE_REC *)GET_LIST_NEXT(triangle_list, t);
  }
  
  for (i = 0; i < n; i++) {
    destroy_edge(er[i]);
  }
  
  return sr;
}


static int
num_of_matches(TRIANGLE_REC *t, int n, EDGE_REC *er[MAX_CONNECTIVITY])
{
  int i, m = 0;
  EDGE_REC *ep;

  for (i = 0; i < n; i++) {
    ep = er[i];
    if (ep == t->edge_1 || ep == t->edge_2 || ep == t->edge_3) m++;
  }
  return m;
}
