
#include "targe2cf.h"

#ifdef X_XFACE

#include "xiface.h"
#include "edges.h"
#include "mctlp.h"

int 
build_xiface(int argc, char *argv[])
{
  
  ESIBuildInterface(ESI_GRAPHIC_EDITOR_MASK, argc, argv);

  EMSetAssocFringeTable(ESIModel(), ColorCreateFringeTable());
  ColorSetupFringeTableByMinMax(EMGetAssocFringeTable(ESIModel()), -1, 1);
        
  ESIPopup();

  ESIHandleCmd("NEWF");
  
  EVSetViewOrientation(EMAnyViewP(ESIModel()), VIEW_ORIENT_TOP);

  return 1;
}


static void 
StopComputation(Widget w, XtPointer text_ptr, XtPointer call_data);
static void 
Proceed(Widget w, XtPointer text_ptr, XtPointer call_data);
static void 
EdgeInfo(Widget w, XtPointer text_ptr, XtPointer call_data);
static void 
MeshControlInfo(Widget w, XtPointer text_ptr, XtPointer call_data);


void
ESICustomize(Widget parent_pane)
{
  Widget stop, proceed, info;
  
  proceed = ESIAddButton("proceed", "PROCEED WITH GENERATION",
                         commandWidgetClass,
                         parent_pane, NULL, 0, Proceed, NULL);
  stop = ESIAddButton("stop", "STOP GENERATION", commandWidgetClass,
                      parent_pane, NULL, 0, StopComputation, NULL);
  info = ESIAddButton("info", "EDGE INFO", commandWidgetClass,
                      parent_pane, NULL, 0, EdgeInfo, NULL);
  info = ESIAddButton("mesh_control_info", "MESH CONTROL INFO",
                      commandWidgetClass,
                      parent_pane, NULL, 0, MeshControlInfo, NULL);
}




static void 
StopComputation(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  ESIEventLoop(1, "Stopped.");
}


static void 
Proceed(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  ESIEventLoopProceed();
}


static void 
PrintEdgeInfo(EView *v_p, caddr_t data_p);


static void 
EdgeInfo(Widget w, XtPointer text_ptr, XtPointer call_data)
{
    EVSetSelectActionFunction(PrintEdgeInfo, NULL, NULL);
    EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
                     NULL, NULL,
                     NULL, NULL,
                     NULL, NULL,
                     NULL, NULL);
    ERptStatusMessage(1, 1, "Select edge(s)");
}


static void 
PrintEdgeInfo(EView *v_p, caddr_t data_p)
{
    GraphicObj *grep = NULL;
    int i;
    FILE *fp;
    char buf[L_tmpnam];
    char s[256];
    
    if (EVItemsInSelectionList() <= 0) {
        ERptErrMessage(1, 2, "No edges selected.", ERROR_GRADE);
        goto exit;
    }
    
    tmpnam(buf);
    fp = fopen(buf, "w");
    fprintf(fp, "\n   EDGE INFO:\n");
    
    i = 0; 
    while ((grep = EVGetFromSelectionList()) != NULL) {
        if (EGGraphicType(grep) == EG_LINE3D) {
            print_edge_data(fp, EGGetAppObj(grep));
        }
    }
    
    fclose(fp);
    ESIPopupInfoFromFile(buf);
    sprintf(s, "rm -f %s", buf);
    system(s);

  exit:
    EVUninstallHandler(v_p);
    ERptStatusMessage(1, 1, "Done");
}


static void 
print_mesh_control_info(EView *v_p, char *file, WCRec *p);

static void 
MeshControlInfo(Widget w, XtPointer text_ptr, XtPointer call_data)
{
    EVSetApplyToViewFunction(print_mesh_control_info, NULL, NULL);
    EMPushHandler(ESIModel(), EVApplyToViewHandler, NULL);
    ERptStatusMessage(1, 1, "Pick a point");
}


static void 
print_mesh_control_info(EView *v_p, char *file, WCRec *p)
{
    char s[256];
    
    sprintf(s, " MESH CONTROL INFO: S=%g, M=%g, A=%g\n",
            get_mesh_cdim(p->x, p->y),
            get_mesh_2nd_dir_magn_factor(p->x, p->y),
            get_mesh_1st_dir_angle(p->x, p->y));
    
    ERptStatusMessage(1, 1, s);
}



#endif /* X_XFACE defined */
